from __future__ import annotations
from negmas import NaiveTitForTatNegotiator, SAOMechanism, make_issue
from negmas.gb.negotiators.micro import MiCRONegotiator
from negmas.preferences import LinearAdditiveUtilityFunction as LUFun
from negmas.preferences.value_fun import AffineFun, IdentityFun, LinearFun, TableFun
from negmas.sao.negotiators.timebased import BoulwareTBNegotiator

SHOW_PLOTS = False


def run_buyer_seller(buyer, seller, normalized=False, callbacks=False, n_steps=100):
    # create negotiation agenda (issues)
    issues = [
        make_issue(name="price", values=10),
        make_issue(name="quantity", values=(1, 11)),
        make_issue(name="delivery_time", values=["today", "tomorrow", "nextweek"]),
    ]

    # create the mechanism
    session = SAOMechanism(
        issues=issues, n_steps=n_steps, time_limit=None, extra_callbacks=callbacks
    )

    # define buyer and seller utilities
    seller_utility = LUFun(
        values=[  # type: ignore
            IdentityFun(),
            LinearFun(0.2),
            TableFun(dict(today=1.0, tomorrow=0.2, nextweek=0.0)),
        ],
        outcome_space=session.outcome_space,
    )

    buyer_utility = LUFun(
        values={  # type: ignore
            "price": AffineFun(-1, bias=9.0),
            "quantity": LinearFun(0.2),
            "delivery_time": TableFun(dict(today=0, tomorrow=0.7, nextweek=1.0)),
        },
        outcome_space=session.outcome_space,
    )
    if normalized:
        seller_utility = seller_utility.scale_max(1.0)
        buyer_utility = buyer_utility.scale_max(1.0)

    # create and add buyer and seller negotiators
    b = buyer(name="buyer")
    s = seller(name="seller")
    s.name += f"{s.short_type_name}"
    b.name += f"{b.short_type_name}"
    session.add(b, ufun=buyer_utility)
    session.add(s, ufun=seller_utility)

    session.run()
    if SHOW_PLOTS:
        session.plot()
        from matplotlib import pyplot as plt

        plt.show()
    return session


def kind_tft(*args, **kwargs):
    kwargs["kindness"] = 0.01
    kwargs["punish"] = False
    return NaiveTitForTatNegotiator(*args, **kwargs)


def test_buy_sell_asp_asp():
    session = run_buyer_seller(BoulwareTBNegotiator, BoulwareTBNegotiator)
    assert session.agreement


# @pytest.mark.skipif(
#     not NEGMAS_RUN_TEMP_FAILING,
#     reason="Not always getting to an greement. This is not a bug necesarily but should be investigated",
# )
# def test_buy_sell_tft_tft():
#     # todo: find out why this is not always getting and agreeming
#     session = run_buyer_seller(kind_tft, NaiveTitForTatNegotiator)
#     # assert session.agreement, f"{session.plot()}{plt.show()}{session.extended_trace}"
#     assert session.agreement, f"{session.extended_trace}"


def test_buy_sell_asp_tft():
    session = run_buyer_seller(BoulwareTBNegotiator, NaiveTitForTatNegotiator)
    assert session.state.agreement


def test_buy_sell_micro():
    session = run_buyer_seller(
        MiCRONegotiator, MiCRONegotiator, callbacks=True, n_steps=10000
    )
    assert session.agreement


if __name__ == "__main__":
    SHOW_PLOTS = True



from __future__ import annotations
import math
import random
from math import isnan

import hypothesis.strategies as st
import numpy as np
import pkg_resources
import pytest
from hypothesis import given, settings
from hypothesis.core import example
from pytest import mark

from negmas.inout import Scenario
from negmas.outcomes import enumerate_issues, issues_from_xml_str, make_issue
from negmas.outcomes.outcome_space import CartesianOutcomeSpace, make_os
from negmas.preferences import (
    AffineUtilityFunction,
    HyperRectangleUtilityFunction,
    LinearAdditiveUtilityFunction,
    LinearUtilityFunction,
    MappingUtilityFunction,
    UtilityFunction,
    pareto_frontier,
)
from negmas.preferences.crisp.const import ConstUtilityFunction
from negmas.preferences.inv_ufun import PresortingInverseUtilityFunction
from negmas.preferences.ops import (
    calc_outcome_distances,
    calc_outcome_optimality,
    calc_reserved_value,
    calc_scenario_stats,
    estimate_max_dist,
    estimate_max_dist_using_outcomes,
    is_rational,
    make_rank_ufun,
    normalize,
    pareto_frontier_bf,
    pareto_frontier_numpy,
    scale_max,
)
from negmas.sao.mechanism import SAOMechanism
from negmas.sao.negotiators import AspirationNegotiator


@mark.parametrize(["n_issues"], [(2,), (3,)])
def test_preferences_range_linear_late_issues(n_issues):
    issues = tuple(make_issue(values=(0.0, 1.0), name=f"i{i}") for i in range(n_issues))
    rs = [(i + 1.0) * random.random() for i in range(n_issues)]
    ufun = LinearUtilityFunction(weights=rs, reserved_value=0.0, issues=issues)
    assert ufun([0.0] * n_issues) == 0.0
    assert ufun([1.0] * n_issues) == sum(rs)
    rng = ufun.minmax(issues=issues)
    assert rng[0] >= 0.0
    assert rng[1] <= sum(rs)


@mark.parametrize(["n_issues"], [(2,), (3,)])
def test_preferences_range_linear(n_issues):
    issues = tuple(make_issue(values=(0.0, 1.0), name=f"i{i}") for i in range(n_issues))
    rs = [(i + 1.0) * random.random() for i in range(n_issues)]
    ufun = LinearUtilityFunction(weights=rs, reserved_value=0.0, issues=issues)
    assert ufun([0.0] * n_issues) == 0.0
    assert ufun([1.0] * n_issues) == sum(rs)
    rng = ufun.minmax()
    assert rng[0] >= 0.0
    assert rng[1] <= sum(rs)


@mark.parametrize(["n_issues"], [(2,), (3,)])
def test_preferences_range_general(n_issues):
    issues = tuple(make_issue(values=(0.0, 1.0), name=f"i{i}") for i in range(n_issues))
    rs = [(i + 1.0) * random.random() for i in range(n_issues)]
    ufun = MappingUtilityFunction(
        mapping=lambda x: sum(r * v for r, v in zip(rs, x)) if x else float("nan")
    )
    assert ufun([0.0] * n_issues) == 0.0
    assert ufun([1.0] * n_issues) == sum(rs)
    rng = ufun.minmax(issues=issues)
    assert rng[0] >= 0.0
    assert rng[1] <= sum(rs)


@given(
    n_outcomes=st.integers(2, 1000),
    n_negotiators=st.integers(2, 5),
    normalized=st.booleans(),
    sort=st.booleans(),
    r0=st.floats(-1.0, 1.0),
    r1=st.floats(-1.0, 1.0),
)
@settings(deadline=60_000)
@example(
    n_outcomes=2, n_negotiators=2, normalized=False, sort=False, r0=0.0, r1=0.5888671875
)
@example(n_outcomes=2, n_negotiators=2, normalized=False, sort=False, r0=0.0, r1=0.0)
def test_calc_outcome_stats(n_outcomes, n_negotiators, normalized, sort, r0, r1):
    def _test_optim_allow_above1(d, outcome, lst):
        if not lst:
            assert isnan(d)
            return
        assert 0 <= d
        if outcome in lst:
            assert abs(1 - d) < 1e-12
            return

    def _test_optim(d, outcome, lst):
        if not lst:
            assert isnan(d)
            return
        assert 0 <= d <= 1
        if outcome in lst:
            assert abs(1 - d) < 1e-12
            return
        assert 1 - d > 1e-12

    def _test_dist(d, outcome, lst):
        if not lst:
            assert isnan(d)
            return
        if outcome in lst:
            assert abs(d) < 1e-12
            return
        assert abs(d) > 1e-12

    np.random.seed(0)
    os = make_os([make_issue(n_outcomes)])
    outcomes = os.enumerate_or_sample()
    utils = np.random.rand(n_negotiators, n_outcomes)
    if not normalized:
        utils *= 100
        r0 *= 100
        r1 *= 100
    ufuns = [
        MappingUtilityFunction(
            dict(zip(outcomes, _)), outcome_space=os, reserved_value=r0 if not i else r1
        )
        for i, _ in enumerate(utils)
    ]
    stats = calc_scenario_stats(tuple(ufuns))
    [tuple(u(_) for u in ufuns) for _ in outcomes]
    mxoverall = estimate_max_dist(ufuns)
    mxpareto = estimate_max_dist_using_outcomes(ufuns, stats.pareto_utils)
    assert mxoverall >= mxpareto
    for outcome in outcomes:
        outils = tuple(u(outcome) for u in ufuns)
        dists = calc_outcome_distances(outils, stats)
        optim_overall = calc_outcome_optimality(dists, stats, max_dist=mxoverall)
        # optim_pareto = calc_outcome_optimality(dists, stats, max_dist=mxpareto)
        # optim_exact = calc_outcome_optimality(dists, stats, outcome_utils=allutils)
        if is_rational(ufuns, outcome):
            _test_optim_allow_above1(
                optim_overall.max_welfare_optimality,
                outcome,
                stats.max_welfare_outcomes,
            )
            for d, o1, lst in zip(
                (
                    dists.pareto_dist,
                    dists.nash_dist,
                    dists.kalai_dist,
                    # dists.max_relative_welfare,
                ),
                (
                    optim_overall.pareto_optimality,
                    optim_overall.nash_optimality,
                    optim_overall.kalai_optimality,
                    # optim_overall.max_relative_welfare_optimality,
                ),
                (
                    stats.pareto_outcomes,
                    stats.nash_outcomes,
                    stats.kalai_outcomes,
                    # stats.max_relative_welfare_outcomes,
                ),
                strict=True,
            ):
                _test_dist(d, outcome, lst)
                _test_optim(o1, outcome, lst)
        else:
            for lst in (
                stats.pareto_outcomes,
                stats.nash_outcomes,
                stats.kalai_outcomes,
                # stats.max_relative_welfare_outcomes,
            ):
                assert outcome not in lst
                for d in (
                    dists.pareto_dist,
                    dists.nash_dist,
                    dists.kalai_dist,
                    # dists.max_relative_welfare,
                ):
                    _test_dist(d, outcome, lst)
                for o1 in (
                    optim_overall.pareto_optimality,
                    optim_overall.nash_optimality,
                    optim_overall.kalai_optimality,
                    # optim_overall.max_relative_welfare_optimality,
                ):
                    _test_optim(o1, outcome, lst)
                _test_optim_allow_above1(
                    optim_overall.max_welfare_optimality,
                    outcome,
                    stats.max_welfare_outcomes,
                )


@given(n_outcomes=st.integers(2, 1000), r0=st.floats(-1.0, 1.0), f=st.floats(0.0, 1.0))
@settings(deadline=60_000)
@example(n_outcomes=2, r0=0.0, f=0.5)
def test_calc_reserved(n_outcomes, r0, f):
    np.random.seed(0)
    os = make_os([make_issue(n_outcomes)])
    outcomes = os.enumerate_or_sample()
    utils = np.random.rand(n_outcomes)
    ufun = MappingUtilityFunction(
        dict(zip(outcomes, utils)), outcome_space=os, reserved_value=r0
    )
    r = calc_reserved_value(ufun, f)
    ufun.reserved_value = r
    nrational = sum(is_rational([ufun], _) for _ in outcomes)
    assert (
        nrational == math.ceil(f * n_outcomes)
    ), f"Got {nrational} outcomes for reserved value {r} of {n_outcomes} outcomes when using fraction {f}"


def test_calc_reserved_fifty_fifty():
    folder_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/FiftyFifty"
    )

    d = Scenario.from_genius_folder(folder_name, ignore_discount=True)
    assert d is not None and d.outcome_space is not None and d.ufuns is not None
    d.normalize()
    n_outcomes = d.outcome_space.cardinality
    outcomes = d.outcome_space.enumerate_or_sample()
    f = 1.0
    for ufun in d.ufuns:
        r = calc_reserved_value(ufun, f)
        ufun.reserved_value = r
        nrational = sum(is_rational([ufun], _) for _ in outcomes)
        assert (
            nrational == math.ceil(f * n_outcomes)
        ), f"Got {nrational} outcomes for reserved value {r} of {n_outcomes} outcomes when using fraction {f}"
        assert r <= 0.0, f"{r=}"


@given(
    n_outcomes=st.integers(2, 1000),
    n_negotiators=st.integers(2, 5),
    normalize=st.booleans(),
    r0=st.floats(-1.0, 1.0),
    r1=st.floats(-1.0, 1.0),
)
@settings(deadline=60_000)
def test_ranks_match_bf(n_outcomes, n_negotiators, normalize, r0, r1):
    np.random.seed(0)
    os = make_os([make_issue(n_outcomes)])
    outcomes = os.enumerate_or_sample()
    utils = np.random.rand(n_negotiators, n_outcomes)
    ufuns = [
        MappingUtilityFunction(
            dict(zip(outcomes, _)), outcome_space=os, reserved_value=r0 if not i else r1
        )
        for i, _ in enumerate(utils)
    ]
    rank_ufuns = [make_rank_ufun(_, normalize=normalize) for _ in ufuns]
    rank_ufuns_bf = [make_rank_ufun(_, normalize=normalize) for _ in ufuns]
    for u, ubf in zip(rank_ufuns, rank_ufuns_bf):
        assert abs(u.reserved_value - ubf.reserved_value) < 1e-10
        for outcome in list(outcomes) + [None]:
            assert abs(u(outcome) - ubf(outcome)) < 1e-12


@given(
    n_outcomes=st.integers(2, 1000),
    n_negotiators=st.integers(2, 5),
    sort=st.booleans(),
    r0=st.floats(-1.0, 1.0),
    r1=st.floats(-1.0, 1.0),
)
@settings(deadline=60_000)
def test_calc_stats_with_ranks(n_outcomes, n_negotiators, sort, r0, r1):
    def _test(x, bf):
        assert len(x) == len(bf), f"stats:{bf}\nglobal:{x}"
        assert len(bf) == len(set(bf)), f"stats:{bf}\nglobal:{x}"
        assert len(x) == len(set(x)), f"stats:{bf}\nglobal:{x}"
        if sort:
            assert all(list(a == b for a, b in zip(x, bf))), f"b:{bf}\nglobal:{x}"
        else:
            assert set(x) == set(bf), f"bf:{bf}\nglobal:{x}"

    np.random.seed(0)
    os = make_os([make_issue(n_outcomes)])
    outcomes = os.enumerate_or_sample()
    utils = np.random.rand(n_negotiators, n_outcomes)
    ufuns = [
        MappingUtilityFunction(
            dict(zip(outcomes, _)), outcome_space=os, reserved_value=r0 if not i else r1
        )
        for i, _ in enumerate(utils)
    ]
    stats = calc_scenario_stats(tuple(ufuns))
    rank_ufuns = [make_rank_ufun(_) for _ in ufuns]
    rank_stats = calc_scenario_stats(rank_ufuns)
    # applying ranking may change the relative order of pareto outcomes (by changing welfare) but not the set
    _test(sorted(stats.pareto_outcomes), sorted(rank_stats.pareto_outcomes))


@given(
    n_outcomes=st.integers(2, 1000),
    n_negotiators=st.integers(2, 5),
    sort=st.booleans(),
    r0=st.floats(-1.0, 1.0),
    r1=st.floats(-1.0, 1.0),
)
@settings(deadline=60_000)
def test_calc_stats(n_outcomes, n_negotiators, sort, r0, r1):
    def _test(x, bf):
        assert len(x) == len(bf), f"stats:{bf}\nglobal:{x}"
        assert len(bf) == len(set(bf)), f"stats:{bf}\nglobal:{x}"
        assert len(x) == len(set(x)), f"stats:{bf}\nglobal:{x}"
        if sort:
            assert all(list(a == b for a, b in zip(x, bf))), f"b:{bf}\nglobal:{x}"
        else:
            assert set(x) == set(bf), f"bf:{bf}\nglobal:{x}"

    np.random.seed(0)
    os = make_os([make_issue(n_outcomes)])
    outcomes = list(os.enumerate_or_sample())
    utils = np.random.rand(n_negotiators, n_outcomes)
    ufuns = [
        MappingUtilityFunction(
            dict(zip(outcomes, _)), outcome_space=os, reserved_value=r0 if not i else r1
        )
        for i, _ in enumerate(utils)
    ]
    stats = calc_scenario_stats(tuple(ufuns))
    bf, bfoutcomes = stats.pareto_utils, stats.pareto_outcomes
    x, xindices = pareto_frontier(ufuns, outcomes, sort_by_welfare=sort)
    xoutcomes = [outcomes[_] for _ in xindices]
    _test(x, bf)
    _test(xoutcomes, bfoutcomes)


@given(
    n_outcomes=st.integers(2, 1000),
    n_negotiators=st.integers(2, 5),
    sort=st.booleans(),
    r0=st.floats(-1.0, 1.0),
    r1=st.floats(-1.0, 1.0),
)
@settings(deadline=60_000)
@example(n_outcomes=4, n_negotiators=2, sort=False, r0=0.0, r1=0.0)
@example(n_outcomes=2, n_negotiators=2, sort=False, r0=0.0, r1=0.0)
def test_mechanism_pareto_frontier_matches_global(
    n_outcomes, n_negotiators, sort, r0, r1
):
    np.random.seed(0)
    os = make_os([make_issue(n_outcomes)])
    outcomes = os.enumerate_or_sample()
    utils = np.random.rand(n_negotiators, n_outcomes)
    m = SAOMechanism(outcome_space=os)
    ufuns = [
        MappingUtilityFunction(
            dict(zip(outcomes, _)), outcome_space=os, reserved_value=r0 if not i else r1
        )
        for i, _ in enumerate(utils)
    ]
    for u in ufuns:
        m.add(AspirationNegotiator(), preferences=u)
    bf, bfoutcomes = m.pareto_frontier(sort_by_welfare=sort)
    # bf = [_[0] for _ in results]
    # bfoutcomes = [_[1] for _ in results]
    outcomes = list(outcomes)
    x, xindices = pareto_frontier(ufuns, outcomes, sort_by_welfare=sort)
    xoutcomes = [outcomes[_] for _ in xindices]
    assert len(x) == len(bf), f"mech:{bf}\nglobal:{x}"
    assert len(bf) == len(set(bf)), f"mech:{bf}\nglobal:{x}"
    assert len(x) == len(set(x)), f"mech:{bf}\nglobal:{x}"
    if sort:
        assert all(list(a == b for a, b in zip(x, bf))), f"b:{bf}\nglobal:{x}"
    else:
        assert set(x) == set(bf), f"bf:{bf}\nglobal:{x}"

    assert len(xoutcomes) == len(bfoutcomes), f"mech:{bfoutcomes}\nglobal:{xoutcomes}"
    assert len(bfoutcomes) == len(
        set(bfoutcomes)
    ), f"mech:{bfoutcomes}\nglobal:{xoutcomes}"
    assert len(xoutcomes) == len(
        set(xoutcomes)
    ), f"mech:{bfoutcomes}\nglobal:{xoutcomes}"
    if sort:
        assert all(
            list(a == b for a, b in zip(xoutcomes, bfoutcomes))
        ), f"bfoutcomes:{bfoutcomes}\nglobal:{xoutcomes}"
    else:
        assert set(xoutcomes) == set(
            bfoutcomes
        ), f"bfoutcomes:{bfoutcomes}\nglobal:{xoutcomes}"


@given(
    n_outcomes=st.integers(2, 1000), n_negotiators=st.integers(2, 5), sort=st.booleans()
)
@settings(deadline=60_000)
def test_pareto_frontier_numpy_matches_bf2(n_outcomes, n_negotiators, sort):
    np.random.seed(0)
    utils = np.random.rand(n_outcomes, n_negotiators)
    bf = list(pareto_frontier_bf(utils, sort_by_welfare=sort))
    x = list(pareto_frontier_numpy(utils, sort_by_welfare=sort))
    assert len(x) == len(bf), f"bf:{bf}\nfast:{x}"
    assert len(bf) == len(set(bf)), f"bf:{bf}\nfast:{x}"
    assert len(x) == len(set(x)), f"bf:{bf}\nfast:{x}"
    if sort:
        assert all(list(a == b for a, b in zip(x, bf))), f"bf:{bf}\nfast:{x}"
    else:
        assert set(x) == set(bf), f"bf:{bf}\nfast:{x}"


@given(
    n_outcomes=st.integers(2, 1000), n_negotiators=st.integers(2, 5), sort=st.booleans()
)
@settings(deadline=60_000)
def test_pareto_frontier_numpy_matches_bf(n_outcomes, n_negotiators, sort):
    np.random.seed(0)
    utils = np.random.rand(n_outcomes, n_negotiators)
    bf = list(pareto_frontier_bf(utils, sort_by_welfare=sort))
    x = list(pareto_frontier_numpy(utils, sort_by_welfare=sort))
    assert len(x) == len(bf), f"bf:{bf}\nfast:{x}"
    assert len(bf) == len(set(bf)), f"bf:{bf}\nfast:{x}"
    assert len(x) == len(set(x)), f"bf:{bf}\nfast:{x}"
    if sort:
        assert all(list(a == b for a, b in zip(x, bf))), f"bf:{bf}\nfast:{x}"
    else:
        assert set(x) == set(bf), f"bf:{bf}\nfast:{x}"


def test_pareto_frontier_does_not_depend_on_order():
    u1 = [
        0.5337723805661662,
        0.8532272031479199,
        0.4781281413197942,
        0.7242899747791032,
        0.3461879818432919,
        0.2608677043479706,
        0.9419131964655383,
        0.29368079952747694,
        0.6093201983562316,
        0.7066918086398718,
    ]
    u2 = [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]
    welfare = [_1 + _2 for _1, _2 in zip(u1, u2)]
    assert welfare.index(max(welfare)) == 3

    f1 = MappingUtilityFunction(lambda o: u1[o[0]] if o else float("nan"))
    f2 = MappingUtilityFunction(lambda o: u2[o[0]] if o else float("nan"))
    assert all(f1((i,)) == u1[i] for i in range(10))
    assert all(f2((i,)) == u2[i] for i in range(10))
    p1, l1 = pareto_frontier(
        [f1, f2], sort_by_welfare=True, outcomes=[(_,) for _ in range(10)]
    )
    p2, l2 = pareto_frontier(
        [f2, f1], sort_by_welfare=True, outcomes=[(_,) for _ in range(10)]
    )

    assert len(p1) == len(p2)
    # assert l2 == list(reversed(l1))

    assert set(l1) == {6, 3}
    assert set(l2) == {6, 3}
    assert set(p1) == {(0.9419131964655383, 0.0), (0.7242899747791032, 1.0)}
    assert set(p2) == {(1.0, 0.7242899747791032), (0.0, 0.9419131964655383)}
    # reverse order of p2
    p2 = [(_[1], _[0]) for _ in p2]
    for a in p1:
        assert a in p2


def test_linear_utility():
    buyer_utility = LinearAdditiveUtilityFunction(
        {  # type: ignore
            "cost": lambda x: -x,
            "number of items": lambda x: 0.5 * x,
            "delivery": {"delivered": 10.0, "not delivered": -2.0},
        },
        issues=[
            make_issue((0.0, 1.0), "cost"),
            make_issue(10, "number of items"),
            make_issue(["delivered", "not delivered"], "delivery"),
        ],
    )
    assert buyer_utility((1.0, 3, "not delivered")) == -1.0 + 0.5 * 3 - 2.0


def test_linear_utility_construction():
    buyer_utility = LinearAdditiveUtilityFunction(
        {  # type: ignore
            "cost": lambda x: -x,
            "number of items": lambda x: 0.5 * x,
            "delivery": {"delivered": 10.0, "not delivered": -2.0},
        },
        issues=[
            make_issue((0.0, 1.0), "cost"),
            make_issue(10, "number of items"),
            make_issue(["delivered", "not delivered"], "delivery"),
        ],
    )
    assert isinstance(buyer_utility, LinearAdditiveUtilityFunction)
    with pytest.raises(ValueError):
        LinearAdditiveUtilityFunction(
            {  # type: ignore
                "cost": lambda x: -x,
                "number of items": lambda x: 0.5 * x,
                "delivery": {"delivered": 10.0, "not delivered": -2.0},
            }
        )


def test_hypervolume_utility():
    f = HyperRectangleUtilityFunction(
        outcome_ranges=[
            None,
            {0: (1.0, 2.0), 1: (1.0, 2.0)},
            {0: (1.4, 2.0), 2: (2.0, 3.0)},
        ],
        utilities=[
            5.0,
            2.0,
            lambda x: 2 * x[2] + x[0] if x is not None else float("nan"),
        ],
    )
    f_ignore_input = HyperRectangleUtilityFunction(
        outcome_ranges=[
            None,
            {0: (1.0, 2.0), 1: (1.0, 2.0)},
            {0: (1.4, 2.0), 2: (2.0, 3.0)},
        ],
        utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]],
        ignore_issues_not_in_input=True,
    )
    f_ignore_failing = HyperRectangleUtilityFunction(
        outcome_ranges=[
            None,
            {0: (1.0, 2.0), 1: (1.0, 2.0)},
            {0: (1.4, 2.0), 2: (2.0, 3.0)},
        ],
        utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]],
        ignore_failing_range_utilities=True,
    )
    f_ignore_both = HyperRectangleUtilityFunction(
        outcome_ranges=[
            None,
            {0: (1.0, 2.0), 1: (1.0, 2.0)},
            {0: (1.4, 2.0), 2: (2.0, 3.0)},
        ],
        utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]],
        ignore_failing_range_utilities=True,
        ignore_issues_not_in_input=True,
    )

    g = HyperRectangleUtilityFunction(
        outcome_ranges=[{0: (1.0, 2.0), 1: (1.0, 2.0)}, {0: (1.4, 2.0), 2: (2.0, 3.0)}],
        utilities=[2.0, lambda x: 2 * x[2] + x[0]],
    )
    g_ignore_input = HyperRectangleUtilityFunction(
        outcome_ranges=[{0: (1.0, 2.0), 1: (1.0, 2.0)}, {0: (1.4, 2.0), 2: (2.0, 3.0)}],
        utilities=[2.0, lambda x: 2 * x[2] + x[0]],
        ignore_issues_not_in_input=True,
    )
    g_ignore_failing = HyperRectangleUtilityFunction(
        outcome_ranges=[{0: (1.0, 2.0), 1: (1.0, 2.0)}, {0: (1.4, 2.0), 2: (2.0, 3.0)}],
        utilities=[2.0, lambda x: 2 * x[2] + x[0]],
        ignore_failing_range_utilities=True,
    )
    g_ignore_both = HyperRectangleUtilityFunction(
        outcome_ranges=[{0: (1.0, 2.0), 1: (1.0, 2.0)}, {0: (1.4, 2.0), 2: (2.0, 3.0)}],
        utilities=[2.0, lambda x: 2 * x[2] + x[0]],
        ignore_failing_range_utilities=True,
        ignore_issues_not_in_input=True,
    )

    funs = [
        g,
        g_ignore_input,
        g_ignore_failing,
        g_ignore_both,
        f,
        f_ignore_input,
        f_ignore_failing,
        f_ignore_both,
    ]
    outcomes = [
        [1.5, 1.5, 2.5],  # belongs to all volumes
        [1.5, 1.5, 1.0],  # belongs to first
        {0: 1.5, 2: 2.5},
        {0: 11.5, 1: 11.5, 2: 12.5},
        [1.5],
        {2: 2.5},
    ]
    expected = [
        [8.5, 8.5, 8.5, 8.5, 13.5, 13.5, 13.5, 13.5],
        [2.0, 2.0, 2.0, 2.0, 7.0, 7.0, 7.0, 7.0],
        [None, 6.5, None, 6.5, None, 11.5, None, 11.5],
        [0.0, 0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 5.0],
        [None, 0.0, None, 0.0, None, 5.0, None, 5.0],
        [None, 0.0, None, 0.0, None, 5.0, None, 5.0],
    ]

    for outcome, expectation in zip(outcomes, expected):
        utilities = [f(outcome) for f in funs]
        for i, (u, e) in enumerate(zip(utilities, expectation)):
            # print(i, utilities, outcome)
            assert e is None and np.isnan(u) or u == e


@mark.parametrize("utype", (LinearUtilityFunction, LinearAdditiveUtilityFunction))
def test_dict_conversion(utype):
    # def test_dict_conversion():
    #     utype = LinearUtilityFunction
    issues = [make_issue(10)]
    u = utype.random(issues=issues, normalized=False)
    d = u.to_dict()
    u2 = utype.from_dict(d)
    for o in enumerate_issues(issues):
        assert abs(u(o) - u2(o)) < 1e-3


def test_inverse_genius_domain():
    with open(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop/Laptop-C-domain.xml"
        )
    ) as ff:
        issues, _ = issues_from_xml_str(ff.read())
    with open(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
        )
    ) as ff:
        u, _ = UtilityFunction.from_xml_str(ff.read(), issues=issues)
    assert u is not None
    inv = PresortingInverseUtilityFunction(u)
    inv.init()
    for i in range(100):
        v = u(inv.one_in((i / 100.0, i / 100.0), normalized=True))
        assert v - 1e-3 <= v <= v + 0.1


def test_random_linear_utils_are_normalized():
    from negmas.preferences import LinearAdditiveUtilityFunction as U2
    from negmas.preferences import LinearUtilityFunction as U1

    eps = 1e-6
    issues = [make_issue(10), make_issue(5), make_issue(2)]

    for U in (U1, U2):
        u = U.random(issues=issues, normalized=True)
        if U == U2:
            assert 1 - eps <= sum(u.weights) <= 1 + eps
        else:
            assert sum(u.weights) <= 1 + eps
        outcomes = enumerate_issues(issues)
        for w in outcomes:
            assert -1e-6 <= u(w) <= 1 + 1e-6, f"{str(u)}"


def _order(u: list[float]):
    return tuple(_[1] for _ in sorted(zip(u, range(len(u)))))


def _check_order(u: list[float]):
    for a, b in zip(u[1:], u[:-1]):
        assert a <= b


def _relative_fraction(u: list[float]):
    return np.asarray([a / b if b > 1e-3 else 0 for a, b in zip(u[1:], u[:-1])])


rngs = [
    (0.0, 1.0),
    (1.0, 2.0),
    # (float("-inf"), 1.0),
    # (0.0, float("inf")),
    # (float("-inf"), float("inf")),
    # # (1.0, 2.0),
    # (float("-inf"), 2.0),
    # (1.0, float("inf")),
]


@given(
    weights=st.lists(st.integers(-10, 10), min_size=2, max_size=2),
    bias=st.floats(min_value=-5.0, max_value=5.0),
    rng=st.sampled_from(rngs),
)
@example(weights=[2, -1], bias=0.0, rng=(0.0, 1.0))
def test_can_normalize_affine_and_linear_ufun(weights, bias, rng):
    issues = [make_issue(10), make_issue(5)]
    ufun = AffineUtilityFunction(weights=weights, bias=bias, issues=issues)
    outcomes = enumerate_issues(issues)
    u1 = [ufun(w) for w in outcomes]

    if (sum(weights) > 1e-6 and rng == (0.0, float("inf"))) or (
        sum(weights) < 1e-6
        and rng[1] > rng[0] + 1e-6
        and rng[0] != float("-inf")
        and rng[1] != float("inf")
    ):
        with pytest.raises(ValueError):
            nfun = ufun.normalize(rng)
        return
    try:
        nfun = ufun.normalize(rng)
    except ValueError as e:
        # todo if the scale is negative, we should raise this exception. I do not now how to expect that correctl yet
        if "scale" in str(e):
            return
        raise e

    assert (
        isinstance(ufun, AffineUtilityFunction)
        or isinstance(ufun, LinearUtilityFunction)
        or isinstance(ufun, ConstUtilityFunction)
    ), "Normalization of ufun of type "
    f"LinearUtilityFunction should generate an IndependentIssuesUFun but we got {type(nfun).__name__}"
    u2 = [nfun(w) for w in outcomes]

    assert (
        max(u2) < rng[1] + 1e-3 and min(u2) > rng[0] - 1e-3
    ), f"Limits are not correct\n{u1}\n{u2}"

    if rng[0] == float("-inf") and rng[1] == float("inf"):
        assert ufun is nfun, "Normalizing with an infinite range should do nothing"

    # if bias == 0.0 and rng[0] == 0.0 and sum(weights) > 1e-5:
    #     scale = [a / b for a, b in zip(u1, u2) if b != 0 and a != 0]
    #     for a, b in zip(u1, u2):
    #         assert (
    #             abs(a - b) < 1e-5 or abs(b) > 1e-5
    #         ), f"zero values are mapped to zero values"
    #         assert (
    #             abs(a - b) < 1e-5 or abs(a) > 1e-5
    #         ), f"zero values are mapped to zero values"
    #     assert max(scale) - min(scale) < 1e-3, f"ufun did not scale uniformly"
    # order1, order2 = _order(u1), _order(u2)
    # assert order1 == order2, f"normalization changed the order of outcomes\n{u1[37:41]}\n{u2[37:41]}"

    if (
        (rng[1] == float("inf") or rng[0] == float("-inf"))
        and abs(bias) < 1e-3
        and sum(weights) > 1e-5
    ):
        relative1 = _relative_fraction(u1)
        relative2 = _relative_fraction(u2)
        assert (
            np.abs(relative1 - relative2).max() < 1e-3
        ), f"One side normalization should not change the result of dividing outcomes\n{u1}\n{u2}"


def test_normalization():
    with open(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop/Laptop-C-domain.xml"
        )
    ) as ff:
        os = CartesianOutcomeSpace.from_xml_str(ff.read())
    issues = os.issues
    outcomes = list(os.enumerate())
    with open(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
        )
    ) as ff:
        u, _ = UtilityFunction.from_xml_str(ff.read(), issues=issues)
    assert u is not None
    assert abs(float(u(("Dell", "60 Gb", "19'' LCD")) - 21.987727736172488)) < 0.000001
    assert abs(float(u(("HP", "80 Gb", "20'' LCD")) - 22.68559475583014)) < 0.000001
    utils = [float(u(_)) for _ in outcomes]
    max_util, min_util = max(utils), min(utils)
    gt_range = dict(
        zip(outcomes, [(_ - min_util) / (max_util - min_util) for _ in utils])
    )
    gt_max = dict(zip(outcomes, [_ / max_util for _ in utils]))

    with open(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
        )
    ) as ff:
        u, _ = UtilityFunction.from_xml_str(ff.read(), issues=issues)
    assert u is not None
    u = normalize(u, to=(0.0, 1.0))
    utils = [float(u(_)) for _ in outcomes]
    max_util, min_util = max(utils), min(utils)
    assert abs(max_util - 1.0) < 0.001
    assert abs(min_util) < 0.001

    for k, v in gt_range.items():
        assert abs(float(u(k)) - v) < 1e-3, f"Failed for {k} got {(u(k))} expected {v}"

    with open(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
        )
    ) as ff:
        u, _ = UtilityFunction.from_xml_str(ff.read(), issues=issues)
    assert u is not None
    u = scale_max(u, 1.0)
    utils = [u(_) for _ in outcomes]
    max_util, min_util = max(utils), min(utils)
    assert abs(max_util - 1.0) < 0.001

    for k, v in gt_max.items():
        assert abs(v - u(k)) < 1e-3, f"Failed for {k} got {(u(k))} expected {v}"


def test_rank_only_ufun_randomize_no_reserve():
    from negmas.preferences import RankOnlyUtilityFunction

    issues = [make_issue((0, 9)), make_issue((1, 5))]
    outcomes = list(make_os(issues).enumerate_or_sample())
    assert len(outcomes) == 10 * 5
    ufun = LinearUtilityFunction(
        weights=[1, 1], issues=issues, reserved_value=float("-inf")
    )
    ro = RankOnlyUtilityFunction(ufun, randomize_equal=True)
    assert isinstance(ro(None), float)
    assert isinstance(ro.reserved_value, float)
    assert ro.reserved_value == ro(None) == float("-inf")
    assert min(ro._mapping.values()) == 0
    assert max(ro._mapping.values()) == 10 * 5 - 1
    assert any(
        ro((0, _)) == 0 for _ in range(1, 6)
    ), f"{[(_, ro(_)) for _ in [None] + outcomes]}"
    assert ro((9, 5)) == 10 * 5 - 1
    mapping = ro.to_mapping_ufun()
    assert mapping.reserved_value == ro(None)
    assert any(
        mapping((0, _)) == 0 for _ in range(1, 6)
    ), f"{[(_, mapping(_)) for _ in [None] + outcomes]}"
    assert mapping((9, 5)) == 10 * 5 - 1


def test_rank_only_ufun_randomize():
    from negmas.preferences import RankOnlyUtilityFunction

    issues = [make_issue((0, 9)), make_issue((1, 5))]
    outcomes = list(make_os(issues).enumerate_or_sample())
    assert len(outcomes) == 10 * 5
    ufun = LinearUtilityFunction(weights=[1, 1], issues=issues, reserved_value=1.5)
    ro = RankOnlyUtilityFunction(ufun, randomize_equal=True)
    assert isinstance(ro(None), int)
    assert isinstance(ro.reserved_value, int)
    assert ro.reserved_value == ro(None)
    assert min(ro._mapping.values()) == 0
    assert max(ro._mapping.values()) == 10 * 5
    assert any(
        ro((0, _)) == 0 for _ in range(1, 6)
    ), f"{[(_, ro(_)) for _ in [None] + outcomes]}"
    assert ro((9, 5)) == 10 * 5
    mapping = ro.to_mapping_ufun()
    assert isinstance(mapping(None), int)
    assert isinstance(mapping.reserved_value, int)
    assert mapping.reserved_value == ro(None)
    assert any(
        mapping((0, _)) == 0 for _ in range(1, 6)
    ), f"{[(_, mapping(_)) for _ in [None] + outcomes]}"
    assert mapping((9, 5)) == 10 * 5


def test_rank_only_ufun_no_randomize():
    from negmas.preferences import RankOnlyUtilityFunction

    issues = [make_issue((0, 9)), make_issue((1, 5))]
    outcomes = list(make_os(issues).enumerate_or_sample())
    assert len(outcomes) == 10 * 5
    ufun = LinearUtilityFunction(weights=[1, 1], issues=issues, reserved_value=0.5)
    ro = RankOnlyUtilityFunction(ufun, randomize_equal=False)
    assert min(ro._mapping.values()) == 0
    assert max(ro._mapping.values()) < 10 * 5
    assert ro((1, 1)) == ro((0, 2))
    assert ro((2, 1)) == ro((1, 2)) == ro((0, 3))
    assert all(
        ro((9, 5)) > ro(_) for _ in outcomes if _ != (9, 5)
    ), f"{[(_, ro(_)) for _ in outcomes]}"
    mapping = ro.to_mapping_ufun()
    assert isinstance(mapping(None), int)
    assert isinstance(mapping.reserved_value, int)
    assert mapping.reserved_value == ro(None)
    assert all(
        mapping((9, 5)) > mapping(_) for _ in outcomes if _ != (9, 5)
    ), f"{[(_, mapping(_)) for _ in outcomes]}"
    assert mapping((1, 1)) == mapping((0, 2))
    assert mapping((2, 1)) == mapping((1, 2)) == mapping((0, 3))


# def test_calc_outcome_stats_example():
#     n_outcomes = 2
#     n_negotiators = 2
#     normalized = False
#     sort = False
#     r0 = 0.0
#     r1 = 1.0
#
#     def _test_optim_allow_above1(d, outcome, lst):
#         if not lst:
#             assert isnan(d)
#             return
#         assert 0 <= d
#         if outcome in lst:
#             assert abs(1 - d) < 1e-12
#             return
#
#     def _test_optim(d, outcome, lst):
#         if not lst:
#             assert isnan(d)
#             return
#         assert 0 <= d <= 1
#         if outcome in lst:
#             assert abs(1 - d) < 1e-12
#             return
#         assert 1 - d > 1e-12
#
#     def _test_dist(d, outcome, lst):
#         if not lst:
#             assert isnan(d)
#             return
#         if outcome in lst:
#             assert abs(d) < 1e-12
#             return
#         assert abs(d) > 1e-12
#
#     np.random.seed(0)
#     os = make_os([make_issue(n_outcomes)])
#     outcomes = os.enumerate_or_sample()
#     utils = np.random.rand(n_negotiators, n_outcomes)
#     if not normalized:
#         utils *= 100
#         r0 *= 100
#         r1 *= 100
#     ufuns = [
#         MappingUtilityFunction(
#             dict(zip(outcomes, _)), outcome_space=os, reserved_value=r0 if not i else r1
#         )
#         for i, _ in enumerate(utils)
#     ]
#     stats = calc_scenario_stats(ufuns)
#     allutils = [tuple(u(_) for u in ufuns) for _ in outcomes]
#     mxoverall = estimate_max_dist(ufuns)
#     mxpareto = estimate_max_dist_using_outcomes(ufuns, stats.pareto_utils)
#     assert mxoverall >= mxpareto
#     for outcome in outcomes:
#         outils = tuple(u(outcome) for u in ufuns)
#         dists = calc_outcome_distances(outils, stats)
#         optim_overall = calc_outcome_optimality(dists, stats, max_dist=mxoverall)
#         # optim_pareto = calc_outcome_optimality(dists, stats, max_dist=mxpareto)
#         # optim_exact = calc_outcome_optimality(dists, stats, outcome_utils=allutils)
#         if is_rational(ufuns, outcome):
#             _test_optim_allow_above1(
#                 optim_overall.max_welfare_optimality,
#                 outcome,
#                 stats.max_welfare_outcomes,
#             )
#             for d, o1, lst in zip(
#                 (
#                     dists.pareto_dist,
#                     dists.nash_dist,
#                     dists.kalai_dist,
#                     # dists.max_relative_welfare,
#                 ),
#                 (
#                     optim_overall.pareto_optimality,
#                     optim_overall.nash_optimality,
#                     optim_overall.kalai_optimality,
#                     # optim_overall.max_relative_welfare_optimality,
#                 ),
#                 (
#                     stats.pareto_outcomes,
#                     stats.nash_outcomes,
#                     stats.kalai_outcomes,
#                     # stats.max_relative_welfare_outcomes,
#                 ),
#                 strict=True,
#             ):
#                 _test_dist(d, outcome, lst)
#                 _test_optim(o1, outcome, lst)
#         else:
#             for lst in (
#                 stats.pareto_outcomes,
#                 stats.nash_outcomes,
#                 stats.kalai_outcomes,
#                 # stats.max_relative_welfare_outcomes,
#             ):
#                 assert outcome not in lst
#                 for d in (
#                     dists.pareto_dist,
#                     dists.nash_dist,
#                     dists.kalai_dist,
#                     # dists.max_relative_welfare,
#                 ):
#                     _test_dist(d, outcome, lst)
#                 for o1 in (
#                     optim_overall.pareto_optimality,
#                     optim_overall.nash_optimality,
#                     optim_overall.kalai_optimality,
#                     # optim_overall.max_relative_welfare_optimality,
#                 ):
#                     _test_optim(o1, outcome, lst)
#                 _test_optim_allow_above1(
#                     optim_overall.max_welfare_optimality,
#                     outcome,
#                     stats.max_welfare_outcomes,
#                 )


if __name__ == "__main__":
    pytest.main(args=[__file__])



from __future__ import annotations
from abc import abstractmethod
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, runtime_checkable

from negmas.common import Distribution, Value
from negmas.outcomes import Outcome, OutcomeSpace
from negmas.protocols import HasMinMax, XmlSerializable

if TYPE_CHECKING:
    from negmas.outcomes.base_issue import Issue

__all__ = [
    "BasePref",
    "Ordinal",
    "CardinalProb",
    "CardinalCrisp",
    "UFun",
    "UFunProb",
    "UFunCrisp",
    "OrdinalRanking",
    "CardinalRanking",
    "HasReservedOutcome",
    "HasReservedValue",
    "HasReservedDistribution",
    "Randomizable",
    "Scalable",
    "Shiftable",
    "PartiallyShiftable",
    "PartiallyScalable",
    "Normalizable",
    "HasRange",
    "InverseUFun",
    "IndIssues",
    "XmlSerializableUFun",
    "SingleIssueFun",
    "MultiIssueFun",
]

X = TypeVar("X", bound="XmlSerializable")


@runtime_checkable
class XmlSerializableUFun(Protocol):
    """Can be serialized to XML format (compatible with GENIUS)"""

    @classmethod
    @abstractmethod
    def from_xml_str(cls: type[X], xml_str: str, **kwargs) -> X:
        """Imports a utility function from a GENIUS XML string.

        Args:

            xml_str (str): The string containing GENIUS style XML utility function definition

        Returns:

            A utility function object (depending on the input file)
        """

    @abstractmethod
    def to_xml_str(self, **kwargs) -> str:
        """Exports a utility function to a well formatted string"""

    def to_genius(self, file_name: PathLike, **kwargs) -> None:
        """
        Exports a utility function to a GENIUS XML file.

        Args:

            file_name (str): File name to export to

        Returns:

            None

        Remarks:
            See ``to_xml_str`` for all the parameters

        """
        file_name = Path(file_name).absolute()
        if file_name.suffix == "":
            file_name = file_name.parent / f"{file_name.stem}.xml"
        with open(file_name, "w") as f:
            f.write(self.to_xml_str(**kwargs))

    @classmethod
    def from_genius(
        cls: type[X], issues: list[Issue], file_name: PathLike, **kwargs
    ) -> X:
        ...

    @abstractmethod
    def xml(self, issues: list[Issue]) -> str:
        ...


@runtime_checkable
class BasePref(Protocol):
    """Base Protcol for all preferences in NegMAS. All Preferences objects implement this interface"""

    @property
    @abstractmethod
    def type(self) -> str:
        """Returns the preferences type."""

    @property
    @abstractmethod
    def base_type(self) -> str:
        """Returns the utility_function base type ignoring discounting and similar wrappings."""

    @abstractmethod
    def is_volatile(self) -> bool:
        """
        Does the utiltiy of an outcome depend on factors outside the negotiation?


        Remarks:
            - A volatile preferences is one that can change even for the same mechanism state due to outside influence
        """

    @abstractmethod
    def is_session_dependent(self) -> bool:
        """
        Does the utiltiy of an outcome depend on the `NegotiatorMechanismInterface`?
        """

    @abstractmethod
    def is_state_dependent(self) -> bool:
        """
        Does the utiltiy of an outcome depend on the negotiation state?
        """

    @abstractmethod
    def is_stationary(self) -> bool:
        """Is the ufun stationary (i.e. utility value of an outcome is a constant)?"""


@runtime_checkable
class HasReservedDistribution(Protocol):
    """In case of disagreement, a value sampled from `reserved_distribution` will be received by the entity"""

    reserved_distribution: Distribution


@runtime_checkable
class HasReservedValue(Protocol):
    """In case of disagreement, `reserved_value` will be received by the entity"""

    reserved_value: float

    @property
    def reserved_distribution(self) -> Distribution:
        ...


@runtime_checkable
class HasReservedOutcome(Protocol):
    """In case of disagreement, the value of `reserved_outcome` will be received by the entity"""

    reserved_outcome: Outcome


@runtime_checkable
class StationaryConvertible(Protocol):
    """Can be converted to stationary Prefereences (i.e. one indepndent of the negotiation session, state or external factors). The conversion is only accurate at the instant it is done"""

    def to_stationary(self):
        ...


@runtime_checkable
class Ordinal(BasePref, Protocol):
    """
    Can be ordered (at least partially)
    """

    @abstractmethod
    def is_not_worse(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            state: The negotiation state at which the comparison is done

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """

    @abstractmethod
    def is_better(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is strictly better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """

    @abstractmethod
    def is_equivalent(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is strictly equivelent than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """

    @abstractmethod
    def is_not_better(self, first: Outcome, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is worse or equivalent than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """

    @abstractmethod
    def is_worse(self, first: Outcome | None, second: Outcome | None) -> bool:
        """
        Compares two offers using the `ufun` returning whether the first is strictly worse than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Remarks:

            - Should raise `ValueError` if the comparison cannot be done

        """


@runtime_checkable
class CardinalProb(Ordinal, Protocol):
    """
    Differences between outcomes are meaningfull but probabilistic.

    Remarks:
        Inheriting from this class adds `is_not_worse` implementation that is
        extremely conservative. It declares that `first` is not worse than `second`
        only if any sample form `first` is ALWAYS not worse than any sample from
        `second`.
    """

    @abstractmethod
    def difference_prob(self, first: Outcome, second: Outcome) -> Distribution:
        """
        Returns a numeric difference between the utility of the two given outcomes
        """


@runtime_checkable
class CardinalCrisp(CardinalProb, Protocol):
    """
    Differences between outcomes are meaningfull and crisp (i.e. real numbers)
    """

    @abstractmethod
    def difference(self, first: Outcome, second: Outcome) -> float:
        """
        Returns a numeric difference between the utility of the two given outcomes
        """


@runtime_checkable
class UFun(CardinalProb, Protocol):
    """Can be called to map an `Outcome` to a `Distribution` or a `float`"""

    def eval(self, offer: Outcome) -> Value:
        """
        Evaluates the ufun without normalization (See `eval_normalized` )
        """
        ...

    def eval_normalized(
        self,
        offer: Outcome | None,
        above_reserve: bool = True,
        expected_limits: bool = True,
    ) -> Value:
        """
        Evaluates the ufun normalizing the result between zero and one

        Args:
            offer (Outcome | None): offer
            above_reserve (bool): If True, zero corresponds to the reserved value not the minimum
            expected_limits (bool): If True, the expectation of the utility limits will be used for normalization instead of the maximum range and minimum lowest limit

        Remarks:
            - If the maximum and the minium are equal, finite and above reserve, will return 1.0.
            - If the maximum and the minium are equal, initinte or below reserve, will return 0.0.
            - For probabilistic ufuns, a distribution will still be returned.
            - The minimum and maximum will be evaluated freshly every time. If they are already caached in the ufun, the cache will be used.

        """
        ...

    @abstractmethod
    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum for the ufun
        """

    def __call__(self, offer: Outcome | None) -> Value:
        ...


T = TypeVar("T", bound="UFunCrisp")


@runtime_checkable
class UFunCrisp(UFun, Protocol):
    """Can be called to map an `Outcome` to a `float`"""

    def eval(self, offer: Outcome) -> float:
        ...

    def to_stationary(self: T) -> T:
        ...

    def __call__(self, offer: Outcome | None) -> float:
        ...


@runtime_checkable
class UFunProb(UFun, Protocol):
    """Can be called to map an `Outcome` to a `Distribution`"""

    @abstractmethod
    def eval(self, offer: Outcome) -> Distribution:
        ...

    def __call__(self, offer: Outcome | None) -> Distribution:
        ...


@runtime_checkable
class OrdinalRanking(Protocol):
    """Outcomes can be ranked. Supports equality"""

    @abstractmethod
    def rank(
        self, outcomes: list[Outcome | None], descending=True
    ) -> list[list[Outcome | None]]:
        """Ranks the given list of outcomes with weights. `None` stands for the null outcome.

        Returns:
            A list of lists of integers giving the outcome index in the input. The list is sorted by utlity value

        """

    @abstractmethod
    def argrank(
        self, outcomes: list[Outcome | None], descending=True
    ) -> list[list[int | None]]:
        """Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:
            A list of lists of integers giving the outcome index in the input. The list is sorted by utlity value

        """


@runtime_checkable
class CardinalRanking(Protocol):
    """Implements ranking of outcomes with meaningful differences (i.e. each rank is given a value and nearer values are more similar)"""

    @abstractmethod
    def rank_with_weights(
        self, outcomes: list[Outcome | None], descending=True
    ) -> list[tuple[tuple[Outcome | None], float]]:
        """
        Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:

            - A list of tuples each with two values:
                - an list of integers giving the index in the input array (outcomes) of an outcome (at the given utility level)
                - the weight of that outcome
            - The list is sorted by weights descendingly

        """

    @abstractmethod
    def argrank_with_weights(
        self, outcomes: list[Outcome | None], descending=True
    ) -> list[tuple[tuple[Outcome | None], float]]:
        """
        Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:

            - A list of tuples each with two values:
                - an list of integers giving the index in the input array (outcomes) of an outcome (at the given utility level)
                - the weight of that outcome
            - The list is sorted by weights descendingly

        """


@runtime_checkable
class InverseUFun(Protocol):
    """Can be used to get one or more outcomes at a given range"""

    ufun: UFun
    initialized: bool

    def __init__(self, ufun: UFun) -> None:
        ...

    def init(self):
        """
        Used to intialize the inverse ufun. Any computationally expensive initialization should be  done here not in the constructor.
        """

    @abstractmethod
    def some(
        self, rng: float | tuple[float, float], normalized: bool, n: int | None = None
    ) -> list[Outcome]:
        """
        Finds a list of outcomes with utilities in the given range.

        Args:
            rng: The range (or single value) of utility values to search for outcomes
            normalized: if `True`, the input `rng` will be understood as ranging from 0-1 (1=max, 0=min) independent of the ufun actual range
            n: The maximum number of outcomes to return

        Remarks:
            - If the ufun outcome space is continuous a sample of outcomes is returned
            - If the ufun outcome space is discrete **all** outcomes in the range are returned
        """

    @abstractmethod
    def one_in(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
        fallback_to_higher: bool = True,
        fallback_to_best: bool = True,
    ) -> Outcome | None:
        """
        Finds an outcmoe with the given utility value.

        Args:
            rng: The range (or single value) of utility values to search for outcomes
            normalized: if `True`, the input `rng` will be understood as ranging from 0-1 (1=max, 0=min) independent of the ufun actual range
            fall_back_to_higher: if `True`, any outcome above the minimum in the range will be returned if nothing can be found in the range
            fall_back_to_best: if `True`, the best outcome will always be offered if no outcome in the given range is found.
        """

    @abstractmethod
    def best_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Finds an outcome with highest utility within the given range

        Args:
            rng: The range (or single value) of utility values to search for outcomes
            normalized: if `True`, the input `rng` will be understood as ranging from 0-1 (1=max, 0=min) independent of the ufun actual range
        """

    @abstractmethod
    def worst_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Finds an outcome with lowest utility within the given range

        Args:
            rng: The range (or single value) of utility values to search for outcomes
            normalized: if `True`, the input `rng` will be understood as ranging from 0-1 (1=max, 0=min) independent of the ufun actual range
        """

    @abstractmethod
    def within_fractions(self, rng: tuple[float, float]) -> list[Outcome]:
        """
        Finds outocmes within the given fractions of utility values. `rng` is always assumed to be normalized between 0-1
        """

    @abstractmethod
    def within_indices(self, rng: tuple[int, int]) -> list[Outcome]:
        """
        Finds outocmes within the given indices with the best at index 0 and the worst at largest index.

        Remarks:
            - Works only for discrete outcome spaces
        """

    @abstractmethod
    def min(self) -> float:
        """
        Finds the minimum utility value that can be returned.

        Remarks:
            - May be different from the minimum of the whole ufun if there is approximation
        """

    @abstractmethod
    def max(self) -> float:
        """
        Finds the maximum utility value that can be returned.

        Remarks:
            - May be different from the maximum of the whole ufun if there is approximation
        """

    @abstractmethod
    def worst(self) -> Outcome:
        """
        Finds the worst  outcome
        """

    @abstractmethod
    def best(self) -> Outcome:
        """
        Finds the best  outcome
        """

    @abstractmethod
    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum utility values that can be returned.

        Remarks:
            These may be different from the results of `ufun.minmax()` as they can be approximate.
        """

    @abstractmethod
    def extreme_outcomes(self) -> tuple[Outcome, Outcome]:
        """
        Finds the worst and best outcomes that can be returned.

        Remarks:
            These may be different from the results of `ufun.extreme_outcomes()` as they can be approximate.
        """

    @abstractmethod
    def __call__(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Calling an inverse ufun directly is equivalent to calling `one_in()`
        """

    def next_worse(self) -> Outcome | None:
        """Returns the rational outcome with utility just below the last one returned from this function"""
        raise NotImplementedError(
            f"next_below is not implemented for {self.__class__.__name__}"
        )

    def next_better(self) -> Outcome | None:
        """Returns the rational outcome with utility just below the last one returned from this function"""
        raise NotImplementedError(
            f"next_above is not implemented for {self.__class__.__name__}"
        )


@runtime_checkable
class HasRange(HasMinMax, UFun, Protocol):
    """Has a defined range of utility values (a minimum and a maximum) and defined best and worst outcomes"""

    def minmax(
        self,
        outcome_space: OutcomeSpace | None = None,
        issues: list[Issue] | None = None,
        outcomes: list[Outcome] | int | None = None,
        max_cardinality=1000,
        above_reserve=False,
    ) -> tuple[float, float]:
        """Finds the range of the given utility function for the given outcomes

        Args:
            self: The utility function
            issues: List of issues (optional)
            outcomes: A collection of outcomes (optional)
            max_cardinality: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not given)
            above_reserve: If given, the minimum and maximum will be set to reserved value if they were less than it.

        Returns:
            (lowest, highest) utilities in that order

        """
        ...

    def extreme_outcomes(
        self,
        outcome_space: OutcomeSpace | None = None,
        issues: list[Issue] | None = None,
        outcomes: list[Outcome] | int | None = None,
        max_cardinality=1000,
    ) -> tuple[Outcome, Outcome]:
        """Finds the best and worst outcomes

        Args:
            ufun: The utility function
            issues: list of issues (optional)
            outcomes: A collection of outcomes (optional)
            max_cardinality: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not
                            given)
        Returns:
            (worst, best) outcomes

        """
        ...

    def max(self) -> Value:
        """
        Returns maximum utility
        """
        ...

    def min(self) -> Value:
        """
        Returns minimum utility
        """
        ...

    def best(self) -> Outcome:
        """
        Returns best outcome
        """
        ...

    def worst(self) -> Outcome:
        """
        Returns worst outcome
        """
        ...


@runtime_checkable
class IndIssues(BasePref, Protocol):
    """The utility value depends on each `Issue` value through a value function that does not depend on any other issue. (i.e. can be modeled as a `LinearAdditiveUtilityFunction`"""

    values: list[Callable[[Any], float]]
    weights: list[float]
    issues: list[Issue]


@runtime_checkable
class Fun(Protocol):
    """A value function mapping values from one or more issues to a real number"""

    @property
    def dim(self) -> int:
        ...

    def minmax(self, input) -> tuple[float, float]:
        ...

    @abstractmethod
    def shift_by(self, offset: float) -> Fun:
        ...

    @abstractmethod
    def scale_by(self, scale: float) -> Fun:
        ...

    def __call__(self, x) -> float:
        ...


@runtime_checkable
class SingleIssueFun(Fun, Protocol):
    """A value function mapping values from a **single** issue to a real number"""

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        ...

    def min(self, input: Issue) -> float:
        ...

    def max(self, input: Issue) -> float:
        ...


@runtime_checkable
class MultiIssueFun(Fun, Protocol):
    """A value function mapping values from **multiple** issues to a real number"""

    @property
    def dim(self) -> int:
        ...

    def minmax(self, input: tuple[Issue, ...]) -> tuple[float, float]:
        ...

    def shift_by(self, offset: float) -> MultiIssueFun:
        ...

    def scale_by(self, scale: float) -> MultiIssueFun:
        ...

    def xml(self, indx: int, issues: list[Issue], bias=0.0) -> str:
        ...

    def __call__(self, x: tuple) -> float:
        ...


@runtime_checkable
class Shiftable(CardinalProb, Protocol):
    """Can be shifted by a constant amount (i.e. utility values are all shifted by this amount)"""

    @abstractmethod
    def shift_by(self, offset: float, shift_reserved=True) -> Shiftable:
        ...

    @abstractmethod
    def shift_min(self, to: float, rng: tuple[float, float] | None = None) -> Shiftable:
        ...

    @abstractmethod
    def shift_max(self, to: float, rng: tuple[float, float] | None = None) -> Shiftable:
        ...


@runtime_checkable
class Scalable(UFun, Protocol):
    """Can be scaled by a constant amount (i.e. utility values are all multiplied by this amount)"""

    @abstractmethod
    def scale_by(self, scale: float, scale_reserved=True) -> Scalable:
        ...

    @abstractmethod
    def scale_min(self, to: float, rng: tuple[float, float] | None = None) -> Scalable:
        ...

    @abstractmethod
    def scale_max(self, to: float, rng: tuple[float, float] | None = None) -> Scalable:
        ...


@runtime_checkable
class PartiallyShiftable(Scalable, Protocol):
    """Can be shifted by a constant amount for a specific part of the outcome space"""

    @abstractmethod
    def shift_min_for(
        self,
        to,
        outcome_space: OutcomeSpace | None = None,
        issues: list[Issue] | None = None,
        outcomes: list[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> PartiallyScalable:
        ...

    @abstractmethod
    def shift_max_for(
        self,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: list[Issue] | None = None,
        outcomes: list[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> PartiallyScalable:
        ...


@runtime_checkable
class PartiallyScalable(Scalable, BasePref, Protocol):
    """Can be scaled by a constant amount for a specific part of the outcome space"""

    @abstractmethod
    def scale_min_for(
        self,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: list[Issue] | None = None,
        outcomes: list[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> PartiallyScalable:
        ...

    @abstractmethod
    def scale_max_for(
        self,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: list[Issue] | None = None,
        outcomes: list[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> PartiallyScalable:
        ...


N = TypeVar("N", bound="Normalizable")


@runtime_checkable
class Normalizable(Shiftable, Scalable, Protocol):
    """Can be normalized to a given range of values (default is 0-1)"""

    @abstractmethod
    def normalize(self: N, to: tuple[float, float] = (0.0, 1.0)) -> N:
        ...


@runtime_checkable
class Randomizable(Protocol):
    """Random Preferences of this type can be created using a `random` method"""

    @classmethod
    @abstractmethod
    def random(
        cls, outcome_space, reserved_value, normalized=True, **kwargs
    ) -> Randomizable:
        """Generates a random ufun of the given type"""



from __future__ import annotations
import json
import math
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Sequence, TypeVar

from negmas import warnings
from negmas.common import Value
from negmas.helpers.prob import Distribution, Real, ScipyDistribution
from negmas.helpers.types import get_full_type_name
from negmas.outcomes import Issue, Outcome, dict2outcome
from negmas.outcomes.common import check_one_at_most, os_or_none
from negmas.outcomes.issue_ops import issues_from_geniusweb_json
from negmas.outcomes.outcome_space import make_os
from negmas.outcomes.protocols import IndependentIssuesOS, OutcomeSpace
from negmas.preferences.value_fun import TableFun
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize
from negmas.warnings import warn_if_slow

from .preferences import Preferences
from .protocols import InverseUFun
from .value_fun import make_fun_from_xml

if TYPE_CHECKING:
    from negmas.preferences import (
        ConstUtilityFunction,
        ProbUtilityFunction,
        UtilityFunction,
        WeightedUtilityFunction,
    )

__all__ = ["BaseUtilityFunction"]


MAX_CARDINALITY = 10_000_000_000
T = TypeVar("T", bound="BaseUtilityFunction")


# PartiallyScalable,
# HasRange,
# HasReservedValue,
# StationaryConvertible,
# OrdinalRanking,
# CardinalRanking,
# BasePref,
class BaseUtilityFunction(Preferences, ABC):
    """
    Base class for all utility functions in negmas
    """

    def __init__(self, *args, reserved_value: float = float("-inf"), **kwargs):
        super().__init__(*args, **kwargs)
        self.reserved_value = reserved_value
        self._cached_inverse: InverseUFun | None = None
        self._cached_inverse_type: type[InverseUFun] | None = None

    @abstractmethod
    def eval(self, offer: Outcome) -> Value:
        ...

    def to_stationary(self: T) -> T:
        raise NotImplementedError(
            f"I do not know how to convert a ufun of type {self.type_name} to a stationary ufun."
        )

    def extreme_outcomes(
        self,
        outcome_space: OutcomeSpace | None = None,
        issues: Iterable[Issue] | None = None,
        outcomes: Iterable[Outcome] | None = None,
        max_cardinality=100_000,
    ) -> tuple[Outcome, Outcome]:
        check_one_at_most(outcome_space, issues, outcomes)
        outcome_space = os_or_none(outcome_space, issues, outcomes)
        if not outcome_space:
            outcome_space = self.outcome_space
        if outcome_space and not outcomes:
            outcomes = outcome_space.enumerate_or_sample(
                max_cardinality=max_cardinality
            )
        if not outcomes:
            raise ValueError("Cannot find outcomes to use for finding extremes")
        mn, mx = float("inf"), float("-inf")
        worst, best = None, None
        warn_if_slow(len(list(outcomes)), "Extreme Outcomes too Slow")
        for o in outcomes:
            u = self(o)
            if u < mn:
                worst, mn = o, u
            if u > mx:
                best, mx = o, u
        if worst is None or best is None:
            raise ValueError(f"Cound not find worst and best outcomes for {self}")
        return worst, best

    def minmax(
        self,
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        max_cardinality=1000,
        above_reserve=False,
    ) -> tuple[float, float]:
        """Finds the range of the given utility function for the given outcomes

        Args:
            self: The utility function
            issues: List of issues (optional)
            outcomes: A collection of outcomes (optional)
            max_cardinality: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not given)
            above_reserve: If given, the minimum and maximum will be set to reserved value if they were less than it.

        Returns:
            (lowest, highest) utilities in that order

        """
        (worst, best) = self.extreme_outcomes(
            outcome_space, issues, outcomes, max_cardinality
        )
        w, b = self(worst), self(best)
        if isinstance(w, Distribution):
            w = w.min
        if isinstance(b, Distribution):
            b = b.max
        if above_reserve:
            r = self.reserved_value
            if r is None:
                return w, b
            if b < r:
                b, w = r, r
            elif w < r:
                w = r
        return w, b

    @property
    def reserved_distribution(self) -> Distribution:
        return ScipyDistribution(type="uniform", loc=self.reserved_value, scale=0.0)

    def max(self) -> Value:
        _, mx = self.minmax()
        return mx

    def min(self) -> Value:
        mn, _ = self.minmax()
        return mn

    def best(self) -> Outcome:
        _, mx = self.extreme_outcomes()
        return mx

    def worst(self) -> Outcome:
        mn, _ = self.extreme_outcomes()
        return mn

    def eval_normalized(
        self,
        offer: Outcome | None,
        above_reserve: bool = True,
        expected_limits: bool = True,
    ) -> Value:
        """
        Evaluates the ufun normalizing the result between zero and one

        Args:
            offer (Outcome | None): offer
            above_reserve (bool): If True, zero corresponds to the reserved value not the minimum
            expected_limits (bool): If True, the expectation of the utility limits will be used for normalization instead of the maximum range and minimum lowest limit

        Remarks:
            - If the maximum and the minium are equal, finite and above reserve, will return 1.0.
            - If the maximum and the minium are equal, initinte or below reserve, will return 0.0.
            - For probabilistic ufuns, a distribution will still be returned.
            - The minimum and maximum will be evaluated freshly every time. If they are already caached in the ufun, the cache will be used.

        """
        r = self.reserved_value
        u = self.eval(offer) if offer else r
        mn, mx = self.minmax()
        if above_reserve:
            if mx < r:
                mx = mn = float("-inf")
            elif mn < r:
                mn = r
        d = mx - mn
        if isinstance(d, Distribution):
            d = float(d) if expected_limits else d.max
        if isinstance(mn, Distribution):
            mn = float(mn) if expected_limits else mn.min
        if d < 1e-5:
            warnings.warn(
                f"Ufun has equal max and min. The outcome will be normalized to zero if they were finite otherwise 1.0: {mn=}, {mx=}, {r=}, {u=}"
            )
            return 1.0 if math.isfinite(mx) else 0.0
        d = 1 / d
        return (u - mn) * d

    def invert(self, inverter: type[InverseUFun] | None = None) -> InverseUFun:
        """
        Inverts the ufun, initializes it and caches the result.
        """
        from .inv_ufun import PresortingInverseUtilityFunction

        if self._cached_inverse and (
            inverter is None or self._cached_inverse_type == inverter
        ):
            return self._cached_inverse
        if inverter is None:
            inverter = PresortingInverseUtilityFunction
        self._cached_inverse_type = inverter
        self._cached_inverse = inverter(self)
        self._cached_inverse.init()
        return self._cached_inverse

    def is_volatile(self) -> bool:
        return True

    def is_session_dependent(self) -> bool:
        return True

    def is_state_dependent(self) -> bool:
        return True

    def scale_by(
        self: T, scale: float, scale_reserved=True
    ) -> WeightedUtilityFunction | T:
        if scale < 0:
            raise ValueError(f"Cannot scale with a negative multiplier ({scale})")
        from negmas.preferences.complex import WeightedUtilityFunction

        r = (scale * self.reserved_value) if scale_reserved else self.reserved_value
        return WeightedUtilityFunction(
            ufuns=[self], weights=[scale], name=self.name, reserved_value=r
        )

    def scale_min_for(
        self: T,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> T:
        if rng is None:
            mn, _ = self.minmax(outcome_space, issues, outcomes)
        else:
            mn, _ = rng
        scale = to / mn
        return self.scale_by(scale)

    def scale_min(self: T, to: float, rng: tuple[float, float] | None = None) -> T:
        return self.scale_min_for(to, outcome_space=self.outcome_space, rng=rng)

    def scale_max_for(
        self: T,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> T:
        if rng is None:
            _, mx = self.minmax(outcome_space, issues, outcomes)
        else:
            _, mx = rng
        scale = to / mx
        return self.scale_by(scale)

    def scale_max(self: T, to: float, rng: tuple[float, float] | None = None) -> T:
        return self.scale_max_for(to, outcome_space=self.outcome_space, rng=rng)

    def normalize_for(
        self: T,
        to: tuple[float, float] = (0.0, 1.0),
        outcome_space: OutcomeSpace | None = None,
    ) -> T | ConstUtilityFunction:
        max_cardinality: int = MAX_CARDINALITY
        if not outcome_space:
            outcome_space = self.outcome_space
        if not outcome_space:
            raise ValueError(
                "Cannot find the outcome-space to normalize for. "
                "You must pass outcome_space, issues or outcomes or have the ufun being constructed with one of them"
            )
        mn, mx = self.minmax(outcome_space, max_cardinality=max_cardinality)

        d = float(mx - mn)
        if d < 1e-7:
            from negmas.preferences.crisp.const import ConstUtilityFunction

            return ConstUtilityFunction(
                0.0 if mx < self.reserved_value else 1.0,
                outcome_space=self.outcome_space,
                name=self.name,
                reserved_value=1.0 if mn < self.reserved_value else 0.0,
            )

        scale = float(to[1] - to[0]) / d

        u = self.scale_by(scale, scale_reserved=True)
        return u.shift_by(to[0] - scale * mn, shift_reserved=True)

    def normalize(
        self: T, to: tuple[float, float] = (0.0, 1.0), normalize_weights: bool = False
    ) -> T | ConstUtilityFunction:
        _ = normalize_weights
        from negmas.preferences import ConstUtilityFunction

        if not self.outcome_space:
            raise ValueError("Cannot normalize a ufun without an outcome-space")
        mn, mx = self.minmax(self.outcome_space, max_cardinality=MAX_CARDINALITY)

        d = float(mx - mn)
        if d < 1e-8:
            return ConstUtilityFunction(
                0.0 if mx < self.reserved_value else 1.0,
                name=self.name,
                reserved_value=1.0 if mn < self.reserved_value else 0.0,
            )

        scale = float(to[1] - to[0]) / d

        # u = self.shift_by(-mn, shift_reserved=True)
        u = self.scale_by(scale, scale_reserved=True)
        return u.shift_by(to[0] - scale * mn, shift_reserved=True)

    def shift_by(
        self: T, offset: float, shift_reserved=True
    ) -> WeightedUtilityFunction | T:
        from negmas.preferences.complex import WeightedUtilityFunction
        from negmas.preferences.crisp.const import ConstUtilityFunction

        r = (self.reserved_value + offset) if shift_reserved else self.reserved_value
        return WeightedUtilityFunction(
            ufuns=[self, ConstUtilityFunction(offset)],
            weights=[1, 1],
            name=self.name,
            reserved_value=r,
        )

    def shift_min_for(
        self: T,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> T:
        if rng is None:
            mn, _ = self.minmax(outcome_space, issues, outcomes)
        else:
            mn, _ = rng
        offset = to - mn
        return self.shift_by(offset)

    def shift_max_for(
        self: T,
        to: float,
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        rng: tuple[float, float] | None = None,
    ) -> T:
        if rng is None:
            _, mx = self.minmax(outcome_space, issues, outcomes)
        else:
            _, mx = rng
        offset = to - mx
        return self.shift_by(offset)

    def _do_rank(self, vals, descending):
        vals = sorted(vals, key=lambda x: x[1], reverse=descending)
        if not vals:
            return []
        ranks = [([vals[0][0]], vals[0][1])]
        for w, v in vals[1:]:
            if v == ranks[-1][1]:
                ranks[-1][0].append(w)
                continue
            ranks.append(([w], v))
        return ranks

    def argrank_with_weights(
        self, outcomes: Sequence[Outcome | None], descending=True
    ) -> list[tuple[list[Outcome | None], float]]:
        """
        Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:

            - A list of tuples each with two values:
                - an list of integers giving the index in the input array (outcomes) of an outcome (at the given utility level)
                - the weight of that outcome
            - The list is sorted by weights descendingly

        """
        vals = zip(range(len(list(outcomes))), (self(_) for _ in outcomes))
        return self._do_rank(vals, descending)

    def argrank(
        self, outcomes: Sequence[Outcome | None], descending=True
    ) -> list[list[Outcome | None]]:
        """
        Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:
            A list of lists of integers giving the outcome index in the input. The list is sorted by utlity value

        """
        ranks = self.argrank_with_weights(outcomes, descending)
        return [_[0] for _ in ranks]

    def rank_with_weights(
        self, outcomes: Sequence[Outcome | None], descending=True
    ) -> list[tuple[list[Outcome | None], float]]:
        """
        Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:

            - A list of tuples each with two values:
                - an list of integers giving the index in the input array (outcomes) of an outcome (at the given utility level)
                - the weight of that outcome
            - The list is sorted by weights descendingly

        """
        vals = zip(outcomes, (self(_) for _ in outcomes))
        return self._do_rank(vals, descending)

    def rank(
        self, outcomes: Sequence[Outcome | None], descending=True
    ) -> list[list[Outcome | None]]:
        """
        Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:
            A list of lists of integers giving the outcome index in the input. The list is sorted by utlity value

        """
        ranks = self.rank_with_weights(outcomes, descending)
        return [_[0] for _ in ranks]

    def eu(self, offer: Outcome | None) -> float:
        """
        calculates the **expected** utility value of the input outcome
        """
        return float(self(offer))

    def to_crisp(self) -> UtilityFunction:
        from negmas.preferences.crisp_ufun import CrispAdapter

        return CrispAdapter(self)

    def to_prob(self) -> ProbUtilityFunction:
        from negmas.preferences.prob_ufun import ProbAdapter

        return ProbAdapter(self)

    def to_dict(self) -> dict[str, Any]:
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        return dict(
            **d,
            outcome_space=serialize(self.outcome_space),
            reserved_value=self.reserved_value,
            name=self.name,
            id=self.id,
        )

    @classmethod
    def from_dict(cls, d):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["outcome_space"] = deserialize(d.get("outcome_space", None))
        return cls(**d)

    def sample_outcome_with_utility(
        self,
        rng: tuple[float, float],
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        n_trials: int = 100,
    ) -> Outcome | None:
        """
        Samples an outcome in the given utiltity range or return None if not possible

        Args:
            rng (Tuple[float, float]): rng
            outcome_space (OutcomeSpace | None): outcome_space
            issues (Sequence[Issue] | None): issues
            outcomes (Sequence[Outcome] | None): outcomes
            n_trials (int): n_trials

        Returns:
            Optional["Outcome"]:
        """
        if rng[0] is None:
            rng = (float("-inf"), rng[1])
        if rng[1] is None:
            rng = (rng[0], float("inf"))
        outcome_space = os_or_none(outcome_space, issues, outcomes)
        if not outcome_space:
            outcome_space = self.outcome_space
        if not outcome_space:
            raise ValueError("No outcome-space is given or defined for the ufun")
        if outcome_space.cardinality < n_trials:
            n_trials = outcome_space.cardinality  # type: ignore I know that it is an int (see the if)
        for o in outcome_space.sample(n_trials, with_replacement=False):
            if o is None:
                continue
            assert (
                o in outcome_space
            ), f"Sampled outcome {o} which is not in the outcome-space {outcome_space}"
            if rng[0] - 1e-6 <= float(self(o)) <= rng[1] + 1e-6:
                return o
        return None

    @classmethod
    def from_xml_str(
        cls,
        xml_str: str,
        issues: Iterable[Issue] | Sequence[Issue],
        safe_parsing=True,
        ignore_discount=False,
        ignore_reserved=False,
        name: str | None = None,
    ) -> tuple[BaseUtilityFunction | None, float | None]:
        """Imports a utility function from a GENIUS XML string.

        Args:

            xml_str (str): The string containing GENIUS style XML utility function definition
            issues (Sequence[Issue] | None): Optional issue space to confirm that the utility function is valid
            product of all issues in the input
            safe_parsing (bool): Turn on extra checks

        Returns:

            A utility function object (depending on the input file)


        Examples:

            >>> from negmas.preferences import UtilityFunction
            >>> import pkg_resources
            >>> from negmas.inout import load_genius_domain
            >>> domain = load_genius_domain(
            ...     pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-domain.xml"
            ...     )
            ... )
            >>> with open(
            ...     pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
            ...     ),
            ...     "r",
            ... ) as ff:
            ...     u, _ = UtilityFunction.from_xml_str(ff.read(), issues=domain.issues)
            >>> with open(
            ...     pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
            ...     ),
            ...     "r",
            ... ) as ff:
            ...     u, _ = UtilityFunction.from_xml_str(ff.read(), issues=domain.issues)
            >>> assert abs(u(("Dell", "60 Gb", "19'' LCD")) - 21.987727736172488) < 0.000001
            >>> assert abs(u(("HP", "80 Gb", "20'' LCD")) - 22.68559475583014) < 0.000001


        """
        from negmas.preferences.complex import WeightedUtilityFunction
        from negmas.preferences.crisp.linear import (
            AffineUtilityFunction,
            LinearAdditiveUtilityFunction,
        )
        from negmas.preferences.crisp.nonlinear import HyperRectangleUtilityFunction

        root = ET.fromstring(xml_str)
        if safe_parsing and root.tag != "utility_space":
            raise ValueError(f"Root tag is {root.tag}: Expected utility_space")

        issues = list(issues)
        ordered_issues: list[Issue] = []
        domain_issues_dict: dict[str, Issue] | None = None
        ordered_issues = issues
        domain_issues_dict = dict(zip([_.name for _ in issues], issues))
        # issue_indices = dict(zip([_.name for _ in issues], range(len(issues))))
        objective = None
        reserved_value = 0.0
        discount_factor = 0.0
        for child in root:
            if child.tag == "objective":
                objective = child
            elif child.tag == "reservation":
                reserved_value = float(child.attrib["value"])
            elif child.tag == "discount_factor":
                discount_factor = float(child.attrib["value"])

        if objective is None:
            objective = root
        weights = {}
        found_issues = {}
        issue_info = {}
        issue_keys = {}
        rects, rect_utils = [], []
        all_numeric = True
        global_bias = 0

        def _get_hyperrects(ufun, max_utility, utiltype=float):
            utype = ufun.attrib.get("type", "none")
            uweight = float(ufun.attrib.get("weight", 1))
            uagg = ufun.attrib.get("aggregation", "sum")
            if uagg != "sum":
                raise ValueError(
                    f"Hypervolumes combined using {uagg} are not supported (only sum is supported)"
                )
            total_util = utiltype(0)
            rects = []
            rect_utils = []
            if utype == "PlainUfun":
                for rect in ufun:
                    util = utiltype(rect.attrib.get("utility", 0))
                    total_util += util if util > 0 else 0
                    ranges = {}
                    rect_utils.append(util * uweight)
                    for r in rect:
                        ii = int(r.attrib["index"]) - 1
                        # key = issue_keys[ii]
                        ranges[ii] = (
                            utiltype(r.attrib["min"]),
                            utiltype(r.attrib["max"]),
                        )
                    rects.append(ranges)
            else:
                raise ValueError(f"Unknown ufun type {utype}")
            total_util = total_util if not max_utility else max_utility
            return rects, rect_utils

        for child in objective:
            if child.tag == "weight":
                indx = int(child.attrib["index"]) - 1
                if indx < 0 or indx >= len(issues):
                    global_bias += float(child.attrib["value"])
                    continue
                weights[issues[indx].name] = float(child.attrib["value"])
            elif child.tag == "utility_function" or child.tag == "utility":
                utility_tag = child
                max_utility = child.attrib.get("maxutility", None)
                if max_utility is not None:
                    max_utility = float(max_utility)
                ufun_found = False
                for ufun in utility_tag:
                    if ufun.tag == "ufun":
                        ufun_found = True
                        _r, _u = _get_hyperrects(ufun, max_utility)
                        rects += _r
                        rect_utils += _u
                if not ufun_found:
                    raise ValueError(
                        "Cannot find ufun tag inside a utility_function tag"
                    )
            elif child.tag == "issue":
                indx = int(child.attrib["index"]) - 1
                issue_key = child.attrib["name"]
                if (
                    domain_issues_dict is not None
                    and issue_key not in domain_issues_dict.keys()
                ):
                    raise ValueError(
                        f"Issue {issue_key} is not in the input issue names ({domain_issues_dict.keys()})"
                    )
                issue_info[issue_key] = {"name": issue_key, "index": indx}
                issue_keys[indx] = issue_key
                info = {"type": "discrete", "etype": "discrete", "vtype": "discrete"}
                for a in ("type", "etype", "vtype"):
                    info[a] = child.attrib.get(a, info[a])
                issue_info[issue_key].update(info)
                mytype = info["type"]
                # vtype = info["vtype"]
                if domain_issues_dict is None:
                    raise ValueError("unknown domain-issue-dict!!!")

                current_issue = domain_issues_dict[issue_key]

                if mytype == "discrete":
                    found_issues[issue_key] = dict()
                    if current_issue.is_continuous():
                        raise ValueError(
                            f"Got a {mytype} issue but expected a continuous valued issue"
                        )
                elif mytype in ("integer", "real"):
                    lower = current_issue.min_value
                    upper = current_issue.max_value
                    lower, upper = (
                        child.attrib.get("lowerbound", lower),
                        child.attrib.get("upperbound", upper),
                    )
                    for rng_child in child:
                        if rng_child.tag == "range":
                            lower, upper = (
                                rng_child.attrib.get("lowerbound", lower),
                                rng_child.attrib.get("upperbound", upper),
                            )
                    if mytype == "integer":
                        if current_issue.is_continuous():
                            raise ValueError(
                                f"Got a {mytype} issue but expected a continuous valued issue"
                            )
                        lower, upper = int(lower), int(upper)  # type: ignore
                    else:
                        lower, upper = float(lower), float(upper)  # type: ignore
                    if (
                        lower < current_issue.min_value
                        or upper > current_issue.max_value
                    ):  # type: ignore
                        raise ValueError(
                            f"Bounds ({lower}, {upper}) are invalid for issue {issue_key} with bounds: "
                            f"{current_issue.values}"
                        )
                else:
                    raise ValueError(f"Unknown type: {mytype}")
                # now we found ranges for range issues and will find values for all issues
                found_values = False
                for item in child:
                    if item.tag == "item":
                        if mytype != "discrete":
                            raise ValueError(
                                f"cannot specify item utilities for not-discrete type: {mytype}"
                            )
                        all_numeric = False
                        item_indx = int(item.attrib["index"]) - 1
                        item_name = item.attrib.get("value", None)
                        if item_name is None:
                            warnings.warn(
                                f"An item without a value at index {item_indx} for issue {issue_key}",
                                warnings.NegmasIOWarning,
                            )
                            continue
                        # may be I do not need this
                        if current_issue.is_integer():
                            item_name = int(item_name)
                        if current_issue.is_float():
                            item_name = float(item_name)
                        if not current_issue.is_valid(item_name):
                            raise ValueError(
                                f"Value {item_name} is not in the domain issue values: "
                                f"{current_issue.values}"
                            )
                        val = item.attrib.get("evaluation", None)
                        if val is None:
                            raise ValueError(
                                f"Item {item_name} of issue {issue_key} has no evaluation attribute!!"
                            )
                        float(val)
                        found_issues[issue_key][item_name] = float(val)
                        found_values = True
                        issue_info[issue_key]["map_type"] = "dict"
                    elif item.tag == "evaluator":
                        _f, _name = make_fun_from_xml(item)
                        found_issues[issue_key] = _f
                        issue_info[issue_key]["map_type"] = _name
                        found_values = True
                if not found_values and issue_key in found_issues.keys():
                    found_issues.pop(issue_key, None)

        # add utilities specified not as hyper-rectangles
        if not all_numeric and all(_.is_numeric() for _ in issues):
            raise ValueError(
                "Some found issues are not numeric but all input issues are"
            )
        u = None
        if len(found_issues) > 0:
            if all_numeric:
                slopes, biases, ws = [], [], []
                for key in (_.name for _ in issues):
                    if key in found_issues:
                        slopes.append(found_issues[key].slope)
                        biases.append(found_issues[key].bias)
                    else:
                        slopes.append(0.0)
                        biases.append(0.0)
                    ws.append(weights.get(key, 1.0))
                bias = 0.0
                for b, w in zip(biases, ws):
                    bias += b * w
                for i, s in enumerate(slopes):
                    ws[i] *= s

                u = AffineUtilityFunction(
                    weights=ws,
                    outcome_space=make_os(ordered_issues),
                    bias=bias + global_bias,
                )
            else:
                u = LinearAdditiveUtilityFunction(
                    values=found_issues,
                    weights=weights,
                    outcome_space=make_os(ordered_issues),
                    bias=global_bias,
                )

        if len(rects) > 0:
            uhyper = HyperRectangleUtilityFunction(
                outcome_ranges=rects,
                utilities=rect_utils,
                name=name,
                outcome_space=make_os(ordered_issues),
                bias=global_bias,
            )
            if u is None:
                u = uhyper
            else:
                u = WeightedUtilityFunction(
                    ufuns=[u, uhyper],
                    weights=[1.0, 1.0],
                    name=name,
                    outcome_space=make_os(ordered_issues),
                )
        if u is None:
            raise ValueError("No issues found")
        if not ignore_reserved:
            u.reserved_value = reserved_value
        u.name = name
        # if not ignore_discount and discount_factor != 0.0:
        #     from negmas.preferences.discounted import ExpDiscountedUFun
        #     u = ExpDiscountedUFun(ufun=u, discount=discount_factor, name=name)
        if ignore_discount:
            discount_factor = None
        return u, discount_factor

    @classmethod
    def from_genius(
        cls, file_name: PathLike | str, **kwargs
    ) -> tuple[BaseUtilityFunction | None, float | None]:
        """Imports a utility function from a GENIUS XML file.

        Args:

            file_name (str): File name to import from

        Returns:

            A utility function object (depending on the input file)


        Examples:

            >>> from negmas.preferences import UtilityFunction
            >>> import pkg_resources
            >>> from negmas.inout import load_genius_domain
            >>> domain = load_genius_domain(
            ...     pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-domain.xml"
            ...     )
            ... )
            >>> u, d = UtilityFunction.from_genius(
            ...     file_name=pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
            ...     ),
            ...     issues=domain.issues,
            ... )
            >>> u.__class__.__name__
            'LinearAdditiveUtilityFunction'
            >>> u.reserved_value
            0.0
            >>> d
            1.0

        Remarks:
            See ``from_xml_str`` for all the parameters

        """
        kwargs["name"] = str(file_name)
        with open(file_name) as f:
            xml_str = f.read()
        return cls.from_xml_str(xml_str=xml_str, **kwargs)

    @classmethod
    def from_geniusweb_json_str(
        cls,
        json_str: str | dict,
        safe_parsing=True,
        issues: Iterable[Issue] | Sequence[Issue] | None = None,
        ignore_discount=False,
        ignore_reserved=False,
        use_reserved_outcome=False,
        name: str | None = None,
    ) -> tuple[BaseUtilityFunction | None, float | None]:
        """Imports a utility function from a GeniusWeb JSON string.

        Args:

            json_str (str): The string containing GENIUS style XML utility function definition
            issues (Sequence[Issue] | None): Optional issue space to confirm that the utility function is valid
            product of all issues in the input
            safe_parsing (bool): Turn on extra checks

        Returns:

            A utility function object (depending on the input file)

        """
        from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction

        _ = safe_parsing

        if isinstance(json_str, str):
            d = json.loads(json_str)
        else:
            d = json_str
        reserved_outcome, discount_factor, u = None, 1.0, None
        if "LinearAdditiveUtilitySpace" in d.keys():
            udict = d["LinearAdditiveUtilitySpace"]
            domain = (
                make_os(issues_from_geniusweb_json(udict["domain"])[0])
                if "domain" in udict.keys()
                else None
            )
            if domain is None and issues is not None:
                domain = make_os(tuple(issues))
            discount_factor = (
                udict.get("discount_factor", 1.0) if not ignore_discount else 1.0
            )
            reserved_value = udict.get("reserved_value", None)
            uname = udict.get("name", name)
            reserved_dict = udict.get("reservationBid", dict()).get("issuevalues", None)
            if reserved_dict and not ignore_reserved:
                reserved_outcome = dict2outcome(reserved_dict, issues=domain.issues)  # type: ignore
            weights = udict.get("issueWeights", None)
            utils = udict.get("issueUtilities", dict())
            values = dict()
            for iname, idict in utils.items():
                vals = idict.get("discreteutils", dict()).get("valueUtilities", dict())
                values[iname] = TableFun(vals)
            u = LinearAdditiveUtilityFunction(
                values=values,
                weights=weights,
                bias=0.0,
                name=uname,
                reserved_outcome=None,
                reserved_value=None,
                outcome_space=domain,
            )
            if not ignore_reserved:
                if reserved_outcome and not use_reserved_outcome:
                    reserved_value = u(reserved_outcome)
                if use_reserved_outcome:
                    u.reserved_outcome = reserved_outcome  # type: ignore
                u.reserved_value = reserved_value  # type: ignore

        return u, discount_factor

    @classmethod
    def from_geniusweb(
        cls, file_name: PathLike | str, **kwargs
    ) -> tuple[BaseUtilityFunction | None, float | None]:
        """Imports a utility function from a GeniusWeb json file.

        Args:

            file_name (str): File name to import from

        Returns:

            A utility function object (depending on the input file)

        Remarks:
            See ``from_geniusweb_json_str`` for all the parameters

        """
        kwargs["name"] = str(file_name)
        with open(file_name) as f:
            xml_str = f.read()
        return cls.from_geniusweb_json_str(json_str=xml_str, **kwargs)

    def to_xml_str(
        self, issues: Iterable[Issue] | None = None, discount_factor=None
    ) -> str:
        """
        Exports a utility function to a well formatted string
        """
        if not hasattr(self, "xml"):
            raise ValueError(
                f"ufun of type {self.__class__.__name__} has no xml() member and cannot be saved to XML string\nThe ufun params: {self.to_dict()}"
            )
        if issues is None:
            if not isinstance(self.outcome_space, IndependentIssuesOS):
                raise ValueError(
                    "Cannot convert to xml because the outcome-space of the ufun is not a cartesian outcome space"
                )
            issues = self.outcome_space.issues
            n_issues = 0
        else:
            issues = list(issues)
            n_issues = len(issues)
        output = (
            f'<utility_space type="any" number_of_issues="{n_issues}">\n'
            f'<objective index="1" etype="objective" type="objective" description="" name="any">\n'
        )

        output += self.xml(issues=issues)  # type: ignore
        if "</objective>" not in output:
            output += "</objective>\n"
            if discount_factor is not None:
                output += f'<discount_factor value="{discount_factor}" />\n'
        if (
            self.reserved_value is not None
            and self.reserved_value != float("-inf")
            and "<reservation value" not in output
        ):
            output += f'<reservation value="{self.reserved_value}" />\n'
        if "</utility_space>" not in output:
            output += "</utility_space>\n"
        return output

    def to_genius(
        self, file_name: PathLike | str, issues: Iterable[Issue] | None = None, **kwargs
    ):
        """
        Exports a utility function to a GENIUS XML file.

        Args:

            file_name (str): File name to export to
            u: utility function
            issues: The issues being considered as defined in the domain

        Returns:

            None


        Examples:

            >>> from negmas.preferences import UtilityFunction
            >>> from negmas.inout import load_genius_domain
            >>> import pkg_resources
            >>> domain = load_genius_domain(
            ...     domain_file_name=pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-domain.xml"
            ...     )
            ... )
            >>> u, d = UtilityFunction.from_genius(
            ...     file_name=pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/Laptop/Laptop-C-prof1.xml"
            ...     ),
            ...     issues=domain.issues,
            ... )
            >>> u.to_genius(
            ...     discount_factor=d,
            ...     file_name=pkg_resources.resource_filename(
            ...         "negmas", resource_name="tests/data/LaptopConv/Laptop-C-prof1.xml"
            ...     ),
            ...     issues=domain.issues,
            ... )

        Remarks:
            See ``to_xml_str`` for all the parameters

        """
        file_name = Path(file_name).absolute()
        if file_name.suffix == "":
            file_name = file_name.parent / f"{file_name.stem}.xml"
        with open(file_name, "w") as f:
            f.write(self.to_xml_str(issues=issues, **kwargs))

    def difference_prob(
        self, first: Outcome | None, second: Outcome | None
    ) -> Distribution:
        """
        Returns a numeric difference between the utility of the two given outcomes
        """
        f, s = self(first), self(second)
        if not isinstance(f, Distribution):
            f = Real(f)
        if not isinstance(s, Distribution):
            s = Real(s)
        return f - s

    def is_not_worse(self, first: Outcome | None, second: Outcome | None) -> bool:
        return self.difference_prob(first, second) >= 0.0

    def difference(self, first: Outcome | None, second: Outcome | None) -> float:
        """
        Returns a numeric difference between the utility of the two given outcomes
        """
        return float(self(first)) - float(self(second))

    def __call__(self, offer: Outcome | None) -> Value:
        """
        Calculate the utility for a given outcome at the given negotiation state.

        Args:
            offer: The offer to be evaluated.


        Remarks:

            - It calls the abstract method `eval` after opationally adjusting the
              outcome type.
            - It is preferred to override eval instead of directly overriding this method
            - You cannot return None from overriden eval() functions but raise an exception (ValueError) if it was
              not possible to calculate the Value.
            - Return a float from your `eval` implementation.
            - Return the reserved value if the offer was None

        Returns:
            The utility of the given outcome
        """
        if offer is None:
            return self.reserved_value  # type: ignore I know that concrete subclasses will be returning the correct type
        return self.eval(offer)


class _FullyStatic:
    """
    Used internally to indicate that the ufun can **NEVER** change due to anything.
    """

    def is_session_dependent(self) -> bool:
        return False

    def is_volatile(self) -> bool:
        return False

    def is_state_dependent(self) -> bool:
        return False

    def is_stationary(self) -> bool:
        return True


class _ExtremelyDynamic:
    """
    Used internally to indicate that the ufun can change due to anything.
    """

    def is_session_dependent(self) -> bool:
        return True

    def is_volatile(self) -> bool:
        return True

    def is_state_dependent(self) -> bool:
        return True

    def is_stationary(self) -> bool:
        return False



# ruff: noqa: F403
from __future__ import annotations
import negmas.gb.components.acceptance as _n
from negmas.gb.components.acceptance import *

__all__ = [_ for _ in _n.__all__ if not _.startswith("GB")]



from __future__ import annotations
import math
import random
import warnings

# from bisect import bisect_left, bisect_right
from typing import Any, Iterable  # , Sequence

import numpy as np
from numpy import floating, integer
from numpy.typing import NDArray

from negmas.outcomes import Outcome
from negmas.warnings import NegmasUnexpectedValueWarning, warn_if_slow

from .base_ufun import BaseUtilityFunction
from .protocols import InverseUFun

__all__ = ["PresortingInverseUtilityFunction", "SamplingInverseUtilityFunction"]

EPS = 1e-6


def index_above_or_equal(a: NDArray, x: Any, lo: int = 0, hi: int | None = None) -> int:
    "Locate the smallest value greater than or equal to x"
    if hi is None:
        hi = len(a)
    i = np.searchsorted(a[lo : hi + 1], x, side="left")
    return i


def index_below_or_equal(a: NDArray, x: Any, lo: int = 0, hi: int | None = None) -> int:
    "Locate the greatest value less than or equal to x"
    if hi is None:
        hi = len(a)
    i = np.searchsorted(a[lo : hi + 1], x, side="right")
    return i - 1


class SamplingInverseUtilityFunction(InverseUFun):
    """
    A utility function inverter that uses sampling.

    Nothing is done during initialization so the fixed cost of this inverter is minimal.
    Nevertheless, each time the system is asked to find an outcome within some range, it uses
    random sampling which is very inefficient and suffers from the curse of dimensionality.
    """

    def __init__(self, ufun: BaseUtilityFunction, max_samples_per_call: int = 10_000):
        self._ufun = ufun
        self.max_samples_per_call = max_samples_per_call
        self._initialized = True

    @property
    def initialized(self):
        return self._initialized

    @property
    def ufun(self):
        return self._ufun

    def init(self):
        pass

    def all(self, rng: float | tuple[float, float]) -> list[Outcome]:
        """
        Finds all outcomes with in the given utility value range

        Args:
            rng: The range. If a value, outcome utilities must match it exactly

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        raise ValueError(
            f"Cannot find all outcomes in a range ({rng}) using a SamplingInverseUtilityFunction. Try a PresortedInverseUtilityFunction"
        )

    def some(
        self, rng: float | tuple[float, float], normalized: bool, n: int | None = None
    ) -> list[Outcome]:
        """
        Finds some outcomes with the given utility value (if discrete, all)

        Args:
            rng: The range. If a value, outcome utilities must match it exactly
            n: The maximum number of outcomes to return

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        if not isinstance(rng, Iterable):
            rng = (rng - EPS, rng + EPS)
        if not n:
            n = self.max_samples_per_call
        else:
            n *= 3
        if not self._ufun.outcome_space:
            return []
        outcomes = list(self._ufun.outcome_space.sample(n, False, False))
        mn, mx = rng
        u = self.ufun.eval_normalized if normalized else self.ufun.eval
        return [_ for _ in outcomes if mn <= u(_) <= mx]

    def worst_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        some = self.some(rng, normalized)
        if not isinstance(rng, Iterable):
            rng = (rng, rng)
        worst_util, worst = float("inf"), None
        for o in some:
            util = self._ufun(o)
            if util < worst_util:
                worst_util, worst = util, o
        return worst

    def best_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        some = self.some(rng, normalized)
        if not isinstance(rng, Iterable):
            rng = (rng, rng)
        best_util, best = float("-inf"), None
        for o in some:
            util = self._ufun(o)
            if util < best_util:
                best_util, best = util, o
        return best

    def one_in(
        self,
        rng: float | tuple[float, float],
        normalized: float,
        fallback_to_higher: bool = True,
        fallback_to_best: bool = True,
    ) -> Outcome | None:
        if not self._ufun.outcome_space:
            return None
        if not isinstance(rng, Iterable):
            rng = (rng - EPS, rng + EPS)
        u = self.ufun.eval_normalized if normalized else self.ufun.eval
        for _ in range(self.max_samples_per_call):
            o = list(self._ufun.outcome_space.sample(1))[0]
            if rng[0] <= u(o) <= rng[1]:
                return o
        if fallback_to_higher and rng[1] < 1 - EPS:
            return self.one_in(
                (rng[0], 1),
                normalized,
                fallback_to_higher=False,
                fallback_to_best=fallback_to_best,
            )
        if fallback_to_best:
            return self._ufun.best()
        return None

    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum utility values that can be returned.

        Remarks:
            These may be different from the results of `ufun.minmax()` as they can be approximate.
        """
        return self.min(), self.max()

    def extreme_outcomes(self) -> tuple[Outcome | None, Outcome | None]:
        """
        Finds the worst and best outcomes that can be returned.

        Remarks:
            These may be different from the results of `ufun.extreme_outcomes()` as they can be approximate.
        """
        return self.worst(), self.best()

    def __call__(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Calling an inverse ufun directly is equivalent to calling `one_in()`
        """
        return self.one_in(rng, normalized)


class PresortingInverseUtilityFunction(InverseUFun):
    """
    A utility function inverter that uses pre-sorting.

    The outcome-space is sampled if it is continuous and enumerated if it is discrete
    during the call to `init()` and an ordered list of outcomes with their utility
    values is then cached.


    Args:
        ufun: The utility function to be inverted
        levels: discretization levels per issue
        max_cache_size: maximum allowed number of outcomes in the resulting inverse
        rational_only: If true, rational outcomes will be sorted but irrational outcomes will not be sorted (should be faster if the reserved value is high)
        n_waypoints: Used to speedup sampling outcomes at given utilities. The larger, the slower init() will be but the faster worst_in() and best_in()
        eps: Absolute difference between utility values to consider them equal (zero or negative to disable).
        rel_eps: Relative difference between utility values to consider them equal (zero or negative to disable).

    Remarks:
        - The actual limit used to judge ufun equality is max(eps, rel_eps * range) where range is the difference between max and min utilities for rational outcomes.
          Set both eps, rel_eps to zero or a negative number to disable cycling through outcomes with equal utilities (or set cycle=False when calling the appropriate function).
    """

    def __init__(
        self,
        ufun: BaseUtilityFunction,
        levels: int = 10,
        max_cache_size: int = 10_000_000_000,
        rational_only: bool = False,
        n_waypints: int = 10,
        eps: float = 1e-12,
        rel_eps: float = 1e-6,
    ):
        self._ufun = ufun
        self.max_cache_size = max_cache_size
        self.levels = levels
        self._initialized = False
        self.outcomes: list[Outcome] = []
        self._last_rational: int = -1
        self.utils: NDArray[floating[Any]] = []  # type: ignore
        self.rational_only = rational_only
        self._waypoints: NDArray[integer[Any]] = []  # type: ignore
        self._waypoint_values: NDArray[floating[Any]] = []  # type: ignore
        self.__nwaypoints = n_waypints
        self._smallest_indx, self._smallest_val = 0, float("inf")
        self._largest_indx, self._largest_val = -1, float("-inf")
        self._last_returned_from_next: int = -1
        self._eps = eps
        self._rel_eps = rel_eps
        self._near_range: dict[int, tuple[int, int]] = dict()

    @property
    def initialized(self):
        return self._initialized

    @property
    def ufun(self):
        return self._ufun

    def reset(self) -> None:
        self._initialized = False
        self.outcomes, self.utils = [], []  # type: ignore
        self._waypoints, self._waypoint_values = [], []  # type: ignore

    def init(self):
        outcome_space = self._ufun.outcome_space
        if outcome_space is None:
            raise ValueError("Cannot find the outcome space.")
        self._worst, self._best = self._ufun.extreme_outcomes()
        self._min, self._max = (
            float(self._ufun(self._worst)),
            float(self._ufun(self._best)),
        )
        self._range = self._max - self._min
        self._offset = self._min / self._range if self._range > EPS else self._min
        for L in range(self.levels, 0, -1):
            n = outcome_space.cardinality_if_discretized(L)
            if n <= self.max_cache_size:
                break
        else:
            raise ValueError(
                f"Cannot discretize keeping cache size at {self.max_cache_size}. Outcome space cardinality is {outcome_space.cardinality}\nOutcome space: {outcome_space}"
            )
        os = outcome_space.to_discrete(levels=L, max_cardinality=self.max_cache_size)
        outcomes = list(os.enumerate_or_sample(max_cardinality=self.max_cache_size))
        utils = [float(self._ufun.eval(_)) for _ in outcomes]
        # x = len(utils)
        warn_if_slow(
            len(utils),
            "Inverting a large utility function",
            lambda x: (x * math.log2(x)) if x else x,
        )
        r = self._ufun.reserved_value
        r = float(r) if r is not None else float("-inf")
        if self.rational_only:
            rational, irrational = [], []
            ur, uir = [], []
            for u, o in zip(utils, outcomes):
                if u >= r:
                    rational.append(o)
                    ur.append(u)
                else:
                    irrational.append(o)
                    uir.append(u)
        else:
            rational, irrational = outcomes, []
            ur, uir = utils, []
        ur, uir = (
            np.asarray(ur, dtype=float).flatten(),
            np.asarray(uir, dtype=float).flatten(),
        )
        indices = np.argsort(ur)
        ur_sorted = ur[indices]
        if len(ur_sorted) > 0:
            eps = max(self._eps, self._rel_eps * (ur_sorted[-1] - ur_sorted[0]))
        else:
            eps = -1
        if eps > 0:
            try:
                n = len(ur_sorted)
                if n >= 2:
                    scaled = np.asarray(ur_sorted / eps, dtype=int)
                    diffs = np.diff(scaled) != 0
                    indexes = np.nonzero(diffs)[0] + 1
                    groups = np.split(scaled, indexes)
                    lengths = np.asarray([_.size for _ in groups], dtype=int)
                    starts = np.hstack((np.asarray([0], dtype=int), indexes))
                    ends = starts + lengths - 1
                    extended = np.nonzero(lengths > 1)[0]
                    starts, ends = starts[extended], ends[extended]
                    for mn, mx in zip(starts, ends):
                        # assert scaled[mn] == scaled[mx], f"{scaled[mn]=}, {scaled[mx]=}"
                        for indx in range(mn, mx + 1):
                            # assert scaled[indx] == scaled[mx], f"{scaled[indx]=}, {scaled[mx]=}"
                            self._near_range[indx] = (mn, mx)
            except Exception:
                pass
            # for indx, current in enumerate(scaled):
            #     mn_indx = indx
            #     for i in range(indx - 1, -1, -1):
            #         u = scaled[i]
            #         if current != scaled[i]:
            #             mn_indx = i + 1
            #             break
            #     mx_indx = indx
            #     for i in range(indx + 1, n):
            #         u = scaled[i]
            #         if current != scaled[i]:
            #             mx_indx = i - 1
            #             break
            #     if mn_indx < mx_indx:
            #         self._near_range[indx] = (mn_indx, mx_indx)

        # ordered_outcomes = sorted(zip(ur, rational, strict=True))
        # if irrational:
        #     ordered_outcomes += list(zip(uir, irrational, strict=True))
        self._initialized = True
        self._last_rational = len(rational) - 1
        # save waypoints within the sorted outcomes list with known utilities.
        # Used to limit the lo, hi limits when doing bisection later
        self.__nwaypoints = min(self.__nwaypoints, self._last_rational + 1)
        waypoints: NDArray[integer[Any]] = (
            np.asarray(
                np.linspace(0, self._last_rational, self.__nwaypoints, endpoint=True),
                dtype=int,
            )
            if self.__nwaypoints > 0
            else np.empty(0, dtype=int)
        )
        waypoint_values: NDArray[floating[Any]] = (
            ur_sorted[waypoints] if self.__nwaypoints > 0 else np.empty(0, dtype=float)
        )
        assert not any(
            _ is None for _ in waypoints
        ), f"{waypoints=}\n{waypoint_values}\n{self._last_rational}"
        assert not any(
            a < b for a, b in zip(waypoints[1:], waypoints[:-1])
        ), f"{waypoints=}\n{waypoint_values}\n{self._last_rational}"
        assert not any(
            a < b for a, b in zip(waypoint_values[1:], waypoint_values[:-1])
        ), f"{waypoints=}\n{waypoint_values}\n{self._last_rational}"
        self._waypoints, self._waypoint_values = waypoints, waypoint_values
        self.outcomes = [rational[_] for _ in indices] + irrational
        self.utils = np.hstack((ur_sorted, uir))
        self._smallest_indx, self._smallest_val = 0, self.utils[0]
        self._largest_indx, self._largest_val = len(self.utils) - 1, self.utils[-1]

    def _un_normalize_range(
        self, rng: float | tuple[float, float], normalized: bool, for_best: bool
    ) -> tuple[float, float]:
        if not isinstance(rng, Iterable):
            rng = (rng - EPS, rng + EPS)
        if not normalized:
            return rng
        mn, mx = self._min, self._max
        d = mx - mn
        if d < EPS:
            return tuple(0.0 if not for_best else 1.0 for _ in rng)
        return tuple(_ * d + mn for _ in rng)

    def _get_limiting_waypoints(self, mn: float, mx: float) -> tuple[int, int]:
        """Returns indices of the smallest utility >= mx and largest utility <= mn in self._utils"""
        if len(self._waypoints) <= 0:
            return (-1, -1)
        lo, hi = self._waypoints[0], self._waypoints[-1]
        lo_val, hi_val = self._waypoint_values[0], self._waypoint_values[-1]
        n = len(self.utils)
        for i, u in zip(self._waypoints, self._waypoint_values):
            if u == lo_val:
                lo = i
            elif u > lo_val:
                lo = max(0, lo - 1)
            if u == hi_val:
                hi = i
            elif u < hi_val:
                hi = min(n - 1, hi - 1)
        # adjust limits to known largest and smallest. This will be specially useful for
        # strategies that call worst_in or best_in repeatedly with descending/ascending values
        if self._smallest_val <= mn and self._smallest_indx > lo:
            lo = self._smallest_indx
        if self._largest_val >= mx and self._largest_indx < hi:
            hi = self._largest_indx
        return lo, hi

    def next_worse(self) -> Outcome | None:
        """Returns the rational outcome with utility just below the last one returned from this function"""
        if self._last_returned_from_next < 0:
            self._last_returned_from_next = self._last_rational
            return self.best()
        if self._last_returned_from_next > 0:
            self._last_returned_from_next -= 1
            return self.outcomes[self._last_returned_from_next]
        return None

    def next_better(self) -> Outcome | None:
        """Returns the rational outcome with utility just above the last one returned from this function"""
        if self._last_returned_from_next < 0:
            self._last_returned_from_next = 0
            return self.worst()
        if self._last_returned_from_next < self._last_rational:
            self._last_returned_from_next += 1
            return self.outcomes[self._last_returned_from_next]
        return None

    def some(
        self, rng: float | tuple[float, float], normalized: bool, n: int | None = None
    ) -> list[Outcome]:
        """
        Finds some outcomes with the given utility value (if discrete, all)

        Args:
            rng: The range. If a value, outcome utilities must match it exactly
            normalized: if given, consider the range as a normalized range betwwen 0 and 1 representing lowest and highest utilities.
            n: The maximum number of outcomes to return

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        if not self._ufun.is_stationary():
            self.init()

        if len(self._waypoints) <= 0:
            return []
        rng = self._un_normalize_range(rng, normalized, True)
        mn, mx = rng
        lo, hi = self._get_limiting_waypoints(mn, mx)
        results = []
        for util, w in zip(self.utils[lo : hi + 1], self.outcomes[lo : hi + 1]):
            if util > mx:
                break
            if util < mn:
                continue
            results.append(w)
        if n and len(results) >= n:
            return random.sample(results, n)
        return results

    def all(self, rng: float | tuple[float, float], normalized: bool) -> list[Outcome]:
        """
        Finds all outcomes with in the given utility value range

        Args:
            rng: The range. If a value, outcome utilities must match it exactly

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        os_ = self._ufun.outcome_space
        if not os_:
            raise ValueError("Unknown outcome space. Cannot invert the ufun")

        if os_.is_discrete():
            return self.some(rng, normalized)
        raise ValueError(
            "Cannot find all outcomes in a range for a continuous outcome space (there is in general an infinite number of them)"
        )

    def _indx_of_worst_in(self, rng: tuple[float, float] | float, normalized: bool):
        if not self._ufun.is_stationary():
            self.init()
        rng = self._un_normalize_range(rng, normalized, False)
        mn, mx = rng
        lo, hi = self._get_limiting_waypoints(mn, mx)
        return index_above_or_equal(self.utils, mn, lo, hi), mn, mx

    def _indx_of_best_in(self, rng: tuple[float, float] | float, normalized: bool):
        if not self._ufun.is_stationary():
            self.init()
        rng = self._un_normalize_range(rng, normalized, True)
        mn, mx = rng
        lo, hi = self._get_limiting_waypoints(mn, mx)
        return index_below_or_equal(self.utils, mx, lo, hi), mn, mx

    def worst_in(
        self, rng: float | tuple[float, float], normalized: bool, cycle: bool = True
    ) -> Outcome | None:
        """
        Finds an outcome with the lowest possible utility within the given range (if any)

        Args:
            rng: Range of utilities to find the outcome within
            normalized: If true, the range will be interpreted as on a scaled ufun with range (0-1)
            cycle: If given, the system will cycle between outcomes with the same utility value (up to `eps`)

        Remarks:
            - Returns None if no such outcome exists, or the worst outcome in the given range is irrational
            - This is an O(log(n)) operation (where n is the number of outcomes)
        """
        indx, mn, mx = self._indx_of_worst_in(rng, normalized)
        if indx > self._last_rational:
            # fail if this worst is irrational
            return None
        if self.utils[indx] > mx:
            # fail if the found outcome is actually worse than the maximum allowed. Should never happen
            return None
        if mn < self._smallest_val:
            self._smallest_indx, self._smallest_val = indx, mn
        if cycle and indx:
            self._cycle_around(indx)
        return self.outcomes[indx]

    def best_in(
        self, rng: float | tuple[float, float], normalized: bool, cycle: bool = True
    ) -> Outcome | None:
        """
        Finds an outcome with the highest possible utility within the given range (if any)

        Args:
            rng: Range of utilities to find the outcome within
            normalized: If true, the range will be interpreted as on a scaled ufun with range (0-1)
            cycle: If given, the system will cycle between outcomes with the same utility value (up to `eps`)

        Remarks:
            - Returns None if no such outcome exists
            - This is an O(log(n)) operation (where n is the number of outcomes)
        """
        indx, mn, mx = self._indx_of_best_in(rng, normalized)
        if indx < 0:
            return None
        if self.utils[indx] < mn:
            return None
        if mx > self._largest_val:
            self._largest_indx, self._largest_val = indx, mx
        if cycle and indx:
            self._cycle_around(indx)
        return self.outcomes[indx]

    def _in(self, x, rng):
        return rng[0] - EPS <= x <= rng[1] + EPS

    def one_in(
        self,
        rng: tuple[float, float],
        normalized: bool,
        fallback_to_higher: bool = True,
        fallback_to_best: bool = True,
    ) -> Outcome | None:
        """
        Finds an outcome within the given range of utilities.

        Remarks:
            - This is an O(log(n)) operation (where n is the number of outcomes)
            - We only search rational outcomes
        """
        mx, rmn, rmx = self._indx_of_best_in(rng, normalized)
        if mx > self._last_rational or not self._in(self.utils[mx], (rmn, rmx)):
            return None
        mn, rmn, rmx = self._indx_of_worst_in(rng, normalized)
        if mn < 0 or not self._in(self.utils[mn], (rmn, rmx)):
            return None
        if mn < 0:
            mn = 0
        if mx > self._last_rational:
            mx = self._last_rational

        def recover_from_failure():
            if fallback_to_higher and rng[1] < 1 - EPS:
                return self.one_in(
                    (rng[0], 1 if normalized else float(self._ufun.max())),
                    normalized,
                    fallback_to_higher=False,
                    fallback_to_best=fallback_to_best,
                )
            if fallback_to_best:
                return self._best
            return None

        if mx < mn:
            # TODO: Something is wrong here. When using stochastic aspiration mn > mx a lot and this leads to strange behavior.
            warnings.warn(
                f"Utility Inverter: {mx=}, {mn=} but we expect mn < mx. Could not find any outcomes in the range given",
                NegmasUnexpectedValueWarning,
            )
            return recover_from_failure()
            # return self.outcomes[min(mx, mn)]
            # mx, mn = mn, mx

        if mn == mx:
            return self.outcomes[mn]
        indx = random.randint(mn, mx)
        return self.outcomes[indx]

    def _cycle_around(self, indx: int) -> None:
        mn_indx, mx_indx = self._near_range.get(indx, (indx, indx))
        if mn_indx < mx_indx:
            np.roll(self.utils[mn_indx : mx_indx + 1], 1)

    def within_fractions(self, rng: tuple[float, float]) -> list[Outcome]:
        """
        Finds outcomes within the given fractions of utility values (the fractions must be between zero and one).

        Remarks:
            - This is an O(n) operation (where n is the number of outcomes)
        """
        if not self._ufun.is_stationary():
            self.init()
        n = self._last_rational + 1
        rng = (rng[0] * n, rng[1] * n)
        rng = (
            max(self._last_rational - math.floor(rng[1]), 0),
            min(n, self._last_rational - math.floor(rng[0])),
        )
        return list(reversed(self.outcomes[int(rng[0]) : int(rng[1]) + 1]))

    def within_indices(self, rng: tuple[int, int]) -> list[Outcome]:
        """
        Finds outcomes within the given indices with the best at index 0 and the worst at largest index.

        Remarks:
            - Works only for discrete outcome spaces
        """
        if not self._ufun.is_stationary():
            self.init()
        n = self._last_rational + 1
        rng = (self._last_rational - rng[1], self._last_rational - rng[0])
        rng = (max(rng[0], 0), min(rng[1], n))
        return list(reversed(self.outcomes[rng[0] : rng[1] + 1]))

    def min(self) -> float:
        """
        Finds the minimum utility value (of the rational outcomes if constructed with `sort_rational_only`)
        """
        if not self._ufun.is_stationary():
            self.init()
        return self.utils[0] if self._last_rational >= 0 else float("inf")

    def max(self) -> float:
        """
        Finds the maximum rational utility value (of the rational outcomes if constructed with `sort_rational_only`)
        """
        if not self._ufun.is_stationary():
            self.init()
        return (
            self.utils[self._last_rational]
            if self._last_rational >= 0
            else float("-inf")
        )

    def worst(self) -> Outcome | None:
        """
        Finds the worst  outcome (of the rational outcomes if constructed with `sort_rational_only`)
        """
        if not self._ufun.is_stationary():
            self.init()
        return self.outcomes[0] if self._last_rational >= 0 else None

    def best(self) -> Outcome | None:
        """
        Finds the best  outcome (of the rational outcomes if constructed with `sort_rational_only`)
        """
        if not self._ufun.is_stationary():
            self.init()
        return self.outcomes[self._last_rational] if self._last_rational >= 0 else None

    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum utility values that can be returned.

        Remarks:

            - These may be different from the results of `ufun.minmax()` as they can be approximate.
            - Will only consider rational outcomes if constructed with `sort_rational_only`
        """
        return self.min(), self.max()

    def extreme_outcomes(self) -> tuple[Outcome | None, Outcome | None]:
        """
        Finds the worst and best outcomes that can be returned.

        Remarks:
            These may be different from the results of `ufun.extreme_outcomes()` as they can be approximate.
        """
        return self.worst(), self.best()

    def __call__(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Calling an inverse ufun directly is equivalent to calling `one_in()`
        """
        return self.one_in(rng, normalized)

    def outcome_at(self, indx: int) -> Outcome | None:
        n = len(self.outcomes)
        if indx >= n:
            return None
        if indx <= self._last_rational:
            return self.outcomes[self._last_rational - indx]
        return self.outcomes[indx]

    def utility_at(self, indx: int) -> float:
        n = len(self.outcomes)
        if indx >= n:
            return float("-inf")
        if indx <= self._last_rational:
            return self.utils[self._last_rational - indx]
        return self.utils[indx]


class PresortingInverseUtilityFunctionBruteForce(InverseUFun):
    """
    A utility function inverter that uses pre-sorting.

    The outcome-space is sampled if it is continuous and enumerated if it is discrete
    during the call to `init()` and an ordered list of outcomes with their utility
    values is then cached.


    Args:
        ufun: The utility function to be inverted
        levels: discretization levels per issue
        max_cache_size: maximum allowed number of outcomes in the resulting inverse
        sort_rational_only: If true, rational outcomes will be sorted but irrational outcomes will not be sorted (should be faster if the reserved value is high)
    """

    def __init__(
        self,
        ufun: BaseUtilityFunction,
        levels: int = 10,
        max_cache_size: int = 10_000_000_000,
        rational_only: bool = False,
    ):
        self._ufun = ufun
        self.max_cache_size = max_cache_size
        self.levels = levels
        self._initialized = False
        self._ordered_outcomes: list[tuple[float, Outcome]] = []
        self.rational_only = rational_only

    @property
    def initialized(self):
        return self._initialized

    @property
    def ufun(self):
        return self._ufun

    def reset(self) -> None:
        self._initialized = False
        self._orddered_outcomes = []

    def init(self):
        outcome_space = self._ufun.outcome_space
        if outcome_space is None:
            raise ValueError("Cannot find the outcome space.")
        self._worst, self._best = self._ufun.extreme_outcomes()
        self._min, self._max = (
            float(self._ufun(self._worst)),
            float(self._ufun(self._best)),
        )
        self._range = self._max - self._min
        self._offset = self._min / self._range if self._range > EPS else self._min
        for L in range(self.levels, 0, -1):
            n = outcome_space.cardinality_if_discretized(L)
            if n <= self.max_cache_size:
                break
        else:
            raise ValueError(
                f"Cannot discretize keeping cache size at {self.max_cache_size}. "
                f"Outcome space cardinality is {outcome_space.cardinality}\nOutcome space: {outcome_space}"
            )
        os = outcome_space.to_discrete(levels=L, max_cardinality=self.max_cache_size)
        outcomes = os.enumerate_or_sample(max_cardinality=self.max_cache_size)
        utils = [float(self._ufun.eval(_)) for _ in outcomes]
        warn_if_slow(
            len(utils),
            "Inverting a large utility function",
            lambda x: (x * math.log2(x)) if x else x,
        )
        r = self._ufun.reserved_value
        r = float(r) if r is not None else float("-inf")
        if self.rational_only:
            rational, irrational = [], []
            ur, uir = [], []
            for u, o in zip(utils, outcomes):
                if u >= r:
                    rational.append(o)
                    ur.append(u)
                else:
                    irrational.append(o)
                    uir.append(u)
        else:
            rational, irrational = list(outcomes), []
            ur, uir = utils, []
        self._ordered_outcomes = sorted(
            zip(ur, rational, strict=True), key=lambda x: -x[0]
        )
        self._last_rational = len(rational) - 1
        if irrational:
            self._ordered_outcomes += list(zip(uir, irrational, strict=True))
        self._initialized = True

    def _normalize_range(
        self, rng: float | tuple[float, float], normalized: bool, for_best: bool
    ) -> tuple[float, float]:
        if not isinstance(rng, Iterable):
            rng = (rng - EPS, rng + EPS)
        if not normalized:
            return rng
        mn, mx = self._min, self._max
        d = mx - mn
        if d < EPS:
            return tuple(1.0 if for_best else 0.0 for _ in rng)
        return tuple(_ * d + mn for _ in rng)

    def some(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
        n: int | None = None,
        fallback_to_higher: bool = True,
        fallback_to_best: bool = True,
    ) -> list[Outcome]:
        """
        Finds some outcomes with the given utility value (if discrete, all)

        Args:
            rng: The range. If a value, outcome utilities must match it exactly
            normalized: if given, consider the range as a normalized range betwwen 0 and 1 representing lowest and highest utilities.
            n: The maximum number of outcomes to return

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        rmx = rng[1] if isinstance(rng, Iterable) else rng
        rmn = rng[0] if isinstance(rng, Iterable) else rng

        def recover_from_failure():
            if fallback_to_higher and rmx < 1 - EPS:
                return self.some(
                    (rmn, 1 if normalized else float(self._ufun.max())),
                    normalized,
                    fallback_to_higher=False,
                    fallback_to_best=fallback_to_best,
                )
            if fallback_to_best:
                return [self._ufun.best()]
            return []

        if not self._ufun.is_stationary():
            self.init()
        rng = self._normalize_range(rng, normalized, True)
        mn, mx = rng
        # todo use bisection
        results = []
        for util, w in self._ordered_outcomes:
            if util > mx:
                continue
            if util < mn:
                break
            results.append(w)
        if n and len(results) >= n:
            return random.sample(results, n)
        if not results:
            return recover_from_failure()
        return results

    def all(self, rng: float | tuple[float, float], normalized: bool) -> list[Outcome]:
        """
        Finds all outcomes with in the given utility value range

        Args:
            rng: The range. If a value, outcome utilities must match it exactly

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first
            - If the outcome-space is discrete, this method will return all outcomes in the given range

        """
        os_ = self._ufun.outcome_space
        if not os_:
            raise ValueError("Unknown outcome space. Cannot invert the ufun")

        if os_.is_discrete():
            return self.some(
                rng, normalized, fallback_to_higher=False, fallback_to_best=False
            )
        raise ValueError(
            "Cannot find all outcomes in a range for a continuous outcome space (there is in general an infinite number of them)"
        )

    def worst_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        if not self._ufun.is_stationary():
            self.init()
        rng = self._normalize_range(rng, normalized, False)
        mn, mx = rng
        if not self._ordered_outcomes:
            return None
        wbefore = None
        for util, w in self._ordered_outcomes[: self._last_rational + 1]:
            if util > mx:
                continue
            if util < mn:
                return wbefore
            wbefore = w
        return wbefore

    def best_in(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        if not self._ufun.is_stationary():
            self.init()
        rng = self._normalize_range(rng, normalized, True)
        mn, mx = rng
        for util, w in self._ordered_outcomes[: self._last_rational + 1]:
            if util > mx:
                continue
            if util < mn:
                return None
            return w
        return None

    def one_in(
        self,
        rng: float | tuple[float, float],
        normalized: bool,
        fallback_to_higher: bool = True,
        fallback_to_best: bool = True,
    ) -> Outcome | None:
        lst = self.some(
            rng,
            normalized,
            fallback_to_higher=fallback_to_higher,
            fallback_to_best=fallback_to_best,
        )

        if not lst:
            return None
        if len(lst) == 1:
            return lst[0]
        return lst[random.randint(0, len(lst) - 1)]

    def within_fractions(self, rng: tuple[float, float]) -> list[Outcome]:
        """
        Finds outcomes within the given fractions of utility values (the fractions must be between zero and one)
        """
        if not self._ufun.is_stationary():
            self.init()
        n = self._last_rational
        rng = (max(rng[0] * n, 0), min(rng[1] * n, n))
        return [_[1] for _ in self._ordered_outcomes[int(rng[0]) : int(rng[1]) + 1]]

    def within_indices(self, rng: tuple[int, int]) -> list[Outcome]:
        """
        Finds outocmes within the given indices with the best at index 0 and the worst at largest index.

        Remarks:
            - Works only for discrete outcome spaces
        """
        if not self._ufun.is_stationary():
            self.init()
        n = self._last_rational + 1
        rng = (max(rng[0], 0), min(rng[1], n))
        return [_[1] for _ in self._ordered_outcomes[rng[0] : rng[1] + 1]]

    def min(self) -> float:
        """
        Finds the minimum utility value
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError("No outcomes to find the best")
        return (
            self._ordered_outcomes[self._last_rational][0]
            if self._last_rational >= 0
            else float("inf")
        )

    def max(self) -> float:
        """
        Finds the maximum utility value
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError("No outcomes to find the best")
        return (
            self._ordered_outcomes[0][0] if self._last_rational >= 0 else float("-inf")
        )

    def worst(self) -> Outcome | None:
        """
        Finds the worst  outcome
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError("No outcomes to find the best")
        return (
            self._ordered_outcomes[self._last_rational][1]
            if self._last_rational >= 0
            else None
        )

    def best(self) -> Outcome | None:
        """
        Finds the best  outcome
        """
        if not self._ufun.is_stationary():
            self.init()
        if not self._ordered_outcomes:
            raise ValueError("No outcomes to find the best")
        return self._ordered_outcomes[0][1] if self._last_rational >= 0 else None

    def minmax(self) -> tuple[float, float]:
        """
        Finds the minimum and maximum utility values that can be returned.

        Remarks:
            These may be different from the results of `ufun.minmax()` as they can be approximate.
        """
        return self.min(), self.max()

    def extreme_outcomes(self) -> tuple[Outcome | None, Outcome | None]:
        """
        Finds the worst and best outcomes that can be returned.

        Remarks:
            These may be different from the results of `ufun.extreme_outcomes()` as they can be approximate.
        """
        return self.worst(), self.best()

    def __call__(
        self, rng: float | tuple[float, float], normalized: bool
    ) -> Outcome | None:
        """
        Calling an inverse ufun directly is equivalent to calling `one_in()`
        """
        return self.one_in(rng, normalized)

    def outcome_at(self, indx: int) -> Outcome | None:
        if indx >= len(self._ordered_outcomes):
            return None
        return self._ordered_outcomes[indx][1]

    def utility_at(self, indx: int) -> float:
        if indx >= len(self._ordered_outcomes):
            return float("-inf")
        return self._ordered_outcomes[indx][0]



from __future__ import annotations
import random
from typing import Any, Generator

from negmas.helpers import unique_name
from negmas.outcomes.base_issue import DiscreteIssue

__all__ = ["CategoricalIssue"]


class CategoricalIssue(DiscreteIssue):
    """
    An `Issue` type representing discrete values that have no ordering or difference defined.
    """

    def __init__(self, values, name=None) -> None:
        super().__init__(values, name)
        values = list(values)
        self._n_values = len(values)
        self._value_type = (
            type(values[0]) if len({type(_) for _ in values}) == 1 else object
        )
        if self.is_numeric():
            self.min_value, self.max_value = min(values), max(values)

    @property
    def type(self) -> str:
        return "categorical"

    def _to_xml_str(self, indx):
        output = f'    <issue etype="discrete" index="{indx + 1}" name="{self.name}" type="discrete" vtype="discrete">\n'

        for i, v in enumerate(self._values):
            output += f'        <item index="{i + 1}" value="{v}" cost="0" description="{v}">\n        </item>\n'
        output += "    </issue>\n"
        return output

    def is_continuous(self) -> bool:
        return False

    def is_uncountable(self) -> bool:
        return False

    @property
    def all(self) -> Generator[Any, None, None]:
        yield from self._values

    def rand_invalid(self):
        """Pick a random *invalid* value"""

        if self.is_float():
            return random.random() * self.max_value + self.max_value * 1.1

        if self.is_integer():
            return random.randint(self.max_value + 1, self.max_value * 2)

        return unique_name("") + str(random.choice(self._values)) + unique_name("")



# -*- coding: utf-8 -*-
"""
Implements Stacked Alternating Offers (SAO) mechanism and basic negotiators.
"""


from .common import *
from .components import *
from .mechanism import *
from .negotiators import *
from .controllers import *

__all__ = (
    common.__all__
    + components.__all__
    + mechanism.__all__
    + negotiators.__all__
    + controllers.__all__
)



from __future__ import annotations
from attrs import define

__all__ = ["Action"]


@define
class Action:
    """An action that an `Agent` can execute in a `World` through the `Simulator`."""

    type: str
    """Action name."""
    params: dict
    """Any extra parameters to be passed for the action."""
